"""
All sources of readable data are described by SqlDataSupplier.  It's getTableStream() method lets you get an
accessor for a table, which is returned as an instance of SqlTableStream.
"""
import functools

from .misc import loose_table_match


class SqlDataSupplier(object):
    """
    A source of data, i.e. tables.
    """
    def getTableStream(self, table_spec: tuple[str]):
        """
        Returns a SqlTableStream for a given table specification, which is a tuple ending in a table name
        and optionally preceded by a database name.
        """


class SqlTableStream(object):
    """
    An iterable set of rows, plus descriptions of the returned fields.
    """
    def __init__(self, upstream=None):
        self.upstream = upstream
        self.context = None

    def supplyContext(self, context):
        """
        Before a compiled object can operate it may need some context.

        :param context:  An instance of SqlEvaluationContext.
        """
        if self.upstream:
            if isinstance(self.upstream, (tuple, list)):
                for s in self.upstream:
                    s.supplyContext(context)
            else:
                self.upstream.supplyContext(context)
        self.context = context

    def getFields(self):
        """
        Return the list of SqlFieldInfo for each field this stream returns.
        """
        if self.upstream:
            return self.upstream.getFields()
        return []

    def getRowIterator(self, where=None, sort=None, start=0, count=-1):
        """
        Return an iterable sequence of rows.  Each row is a tuple containing values for each field listed
        by getFields().
        """
        if self.upstream:
            return self.upstream.getRowIterator(where, sort, start, count)
        return []

    def asDataFrame(self, where=None):
        """
        Convert a TableStream into a pandas DataFrame.
        """
        import pandas
        cols = list(map(lambda f:f.field,self.getFields()))
        cols_out = {col: [] for col in cols}
        for row_data in self.getRowIterator(where):
            for n_col, col in enumerate(cols):
                cols_out[col].append(row_data[n_col])
        return pandas.DataFrame(data=cols_out)

    def canSort(self, sort):
        """
        Implementations can declare whether they support sorting, with a given sort specification.
        """
        if self.upstream:
            return self.upstream.canSort(sort)
        return True

    def canWhere(self, where):
        """
        Implementations can declare whether they can handle a given WHERE clause.
        """
        if self.upstream:
            return self.upstream.canWhere(where)
        return True

    def canLimit(self, start, count):
        """
        Declares whether this implementation can handle the supplied start/count (limit) values.
        """
        if self.upstream:
            return self.upstream.canLimit(start, count)
        return True


class SqlFieldInfo(object):
    """
    Metadata about a single field.
    """
    def __init__(self, table=None, field=None, database=None, alias=None, table_alias=None):
        self.database = database
        self.table = table
        self.table_alias = table_alias
        self.field = field
        self.alias = alias

    def __str__(self):
        out = ""
        if self.database:
            out += self.database + "."
        if self.table:
            out += self.table
        if self.table_alias:
            out += "(" + self.table_alias + ")"
        out += "."
        if self.field:
            out += self.field
        if self.alias:
            out += "(" + self.alias + ")"
        return out


class MemoryDataSupplier(SqlDataSupplier):
    """
    Provides table data from data in memory.
    """
    def __init__(self, data, db_name=None, allow_sort: bool=True):
        self.data = data
        self.dbName = db_name
        self._allow_sort = allow_sort

    def getTableStream(self, table_spec):
        table_name = table_spec[-1]
        db_name = table_spec[-2] if len(table_spec) > 1 else None
        if self.dbName or db_name:
            if not self.dbName or not db_name:
                return None
            if db_name.lower() != self.dbName.lower():
                return None
        real_table_name = loose_table_match(table_name, self.data)
        if not real_table_name:
            return None
        col_names = set()
        rows = self.data[real_table_name]
        for row in rows:
            for col in row.keys():
                col_names.add(col)
        col_names = sorted(list(col_names))
        return MemStream(real_table_name, rows, col_names, table_name=table_name, allow_sort=self._allow_sort)


class MemStream(SqlTableStream):
    def __init__(self, name, rows, col_names, table_name, allow_sort):
        super(MemStream, self).__init__()
        self.rows = rows
        self.fields = [SqlFieldInfo(name, field) for field in col_names]
        self.table_name = table_name
        self._allow_sort = allow_sort

    def supplyContext(self, context):
        self.context = context.derive(fields=self.getFields())

    def getFields(self):
        return self.fields

    def getRowIterator(self, where=None, sort=None, start=0, count=-1):
        rows = (tuple([row.get(field.field) for field in self.fields]) for row in self.rows)
        if sort:
            def sort_cmp(a, b):
                return sort.compare(a, b, self.context)
            rows = sorted(rows, key=functools.cmp_to_key(sort_cmp))
        if start or count >= 0:
            rows = list(rows)
            if count >= 0:
                rows = rows[start:start+count]
            else:
                rows = rows[start:]
        for row in rows:
            if self.context and self.context.debugLevel:
                print("MemoryDataSupplier: t={0}, row:{1}".format(self.table_name, row))
            yield row

    def canLimit(self, start, count):
        return True

    def canWhere(self, where):
        return not where

    def canSort(self, sort):
        return not sort or self._allow_sort



import pandas
import dask
import numpy
import re
import io
import openpyxl
import dslibrary
from dslibrary.utils.filechunker import slicer

from .df_iterator import DFIterator
from .exceptions import SqlException, SqlException_SqlProblem
from .misc import table_name_from_table_spec, format_options_for_table
from .modify_sql import embed_parameters
from .parser import find_referenced_tables
from .s2p_errs import translate_pandas_dask_exceptions
from .sql_to_pandas import SqlToPandas
from .streaming_query import streaming_query, ChunkedRowIterator


def file_sql_read(
        open_table_stream: callable, sql: str, parameters: list=None,
        max_rows: int=0, memory_threshold__data_size=None, memory_threshold__max_rows=None,
        mode: str="auto", format_options=None):
    """
    Run SQL against file-like objects containing columnar data in CSV and other formats.

    :param open_table_stream: Function that will open a file-like stream for table data.
    :param sql:          SQL to run.
    :param parameters:   To fill into sql where '%s' occurs.
    :param max_rows:     Maximum # rows to read from files.
    :param memory_threshold__data_size: If data to process is below this size, process in memory.  Measures sum of file sizes.
    :param memory_threshold__max_rows:  If max_rows is below this size, process in memory.
    :param mode:         'pandas', 'stream', 'dask' or 'auto'.
                            'pandas' = load into pandas.DataFrame, e.g. with read_csv().
                            'stream' = do all analysis 'on the fly', using little memory
                            'dask' = use dask instead of pandas - NOTE: dask does not restrain its memory use in many cases
                            'auto' = decide automatically
    :param format_options: Determines file format for a given table.  A mapping from table name to DataFileMetrics options.  Use '' for default options.
    :return:   A tuple with 0) column names and 1) a row iterator.
    """
    sql = embed_parameters(sql, parameters)
    # make automated decision for 'mode'
    if mode == "auto":
        memory_threshold__max_rows = memory_threshold__max_rows or 10000
        memory_threshold__data_size = memory_threshold__data_size or 50000000
        if max_rows and max_rows < memory_threshold__max_rows:
            # low max_rows: we are only looking at the first part of each table
            mode = "pandas"
        else:
            total_size = total_table_stream_sizes(open_table_stream, sql=sql, external_data_indicator=memory_threshold__data_size)
            if total_size < memory_threshold__data_size:
                # total size of all tables is low enough
                mode = "pandas"
            else:
                # large operation, can't run in memory
                mode = "stream"
    if mode == "stream":
        try:
            return streaming_query(sql, open_table_stream, format_options=format_options)
        except Exception as exc:
            raise translate_pandas_dask_exceptions(exc)
    # do in-memory query
    def table_loader(table_spec):
        table_name = table_name_from_table_spec(table_spec)
        table_format = format_options_for_table(format_options, table_spec)
        if max_rows:
            table_format["nrows"] = max_rows
        clean_spec = (table_name,) if isinstance(table_spec, str) else table_spec[:-1] + (table_name,)
        stream = open_table_stream(clean_spec)
        # TODO this metadata might be useful - it is generated in api_sandbox.open_sandbox_file() and other places
        #  where we generate stream classes
        #secret_metadata = (stream.metadata if hasattr(stream, "metadata") else None) or {}
        try:
            return dslibrary.load_dataframe(stream, format_options=table_format, fallback_to_text=True)
        finally:
            stream.close()
        # NOTE: we're not using dask at the moment but if we wanted dslibrary to choose an engine based on file size...
        # , dask=memory_threshold__data_size
    lookup = DbDictLookup(table_loader)
    db = DynamicDict(lookup.tables, lookup.lookup_table)
    return _run_select_s2p(sql, db)


def data_sql_read(data: (str, bytes), sql: str, parameters: list=None, max_rows=None, format_options: dict=None):
    """
    Query raw data as SQL.
    :param data:      Data with columnar data, i.e. CSV, etc.
    :param sql:       SQL to run against the data.
    :param parameters: Parameter values to fill in where '%s' occurs in the SQL.
    :param max_rows:  Limits the amount of raw input data that will be scanned, not the number of output rows.
    :param format_options: See DataFileMetrics.
    """
    def opener(table_spec):
        return io.StringIO(data) if isinstance(data, str) else io.BytesIO(data)
    return file_sql_read(opener, sql=sql, parameters=parameters, max_rows=max_rows, format_options={"": format_options})


def dataframe_sql_read(sql, df):
    """
    Query against a dataframe.
    :param sql:     SQL to run.
    :param df:      A pandas DataFrame, or (cols, rows), or {col: vals, ...}
    :return:  cols, rows
    """
    if isinstance(df, (tuple, list)):
        df = pandas.DataFrame(data=df[1], columns=df[0])
    elif isinstance(df, dict):
        df = pandas.DataFrame(data=df)
    tables = DynamicDict({"table": df}, lambda t: "table")
    return _run_select_s2p(sql, tables)


class DbDictLookup(object):
    def __init__(self, table_loader):
        """
        Generate a table cache and lookup mechanism, given a table loading function.

        :param table_loader:  Does loose matching and retrieves a table, given a table name or
            table specification.
        """
        self.table_loader = table_loader
        self.tables = {}

    def lookup_table(self, table_ref):
        """
        Load a given table, performing loose matching.
        :param table_ref:  Either a str with a table name, or a tuple with a table spec, i.e.
            ("database", "table").
        :returns:       The key under which this table has been loaded, or None if not found.
        """
        table = self.table_loader(table_ref)
        if table is None:
            return
        table_name = table_ref if isinstance(table_ref, str) else table_ref[-1]
        self.tables[table_name] = table
        return table_name


class DynamicDict(object):
    """
    A {} that looks up un-found elements as needed.
    """
    def __init__(self, initial, lookup):
        self.data = {} if initial is None else initial
        self.lookup = lookup

    def get(self, item, default=None):
        found = self.data.get(item)
        if found is not None:
            return found
        new_index = self.lookup(item)
        if new_index is not None:
            return self.data[new_index]
        return default

    def __getitem__(self, item):
        found = self.get(item)
        if found is None:
            raise SqlException_SqlProblem("Table not found: %s" % str(item))
        return found

    def __contains__(self, item):
        return self.get(item) is not None

    def items(self):
        return self.data.items()


def _run_select_s2p(sql, db):
    # references to 'tables' import each table
    ext_tables = lambda rq: (None, "tables[%s]" % repr(rq))
    # capture simple limit - it's a little faster to implement it outside of the main dask loop than with 'head()'
    ext_proc = {"post_limit": None}
    try:
        # compile the SQL
        sql_pandas_code = SqlToPandas(
            sql, local_tables="tables", ext_table_access=ext_tables, external_processing=ext_proc
        ).code
        # execute, filling in tables and dependencies
        vars = {"tables": db, "pandas": pandas, "dask": dask, "numpy": numpy}
        eval(compile(sql_pandas_code, "<sql>", "exec"), vars, vars)
    except Exception as err:
        raise _translate_s2p_exception(err, sql)
    results_df = vars["_"]
    cols = list(results_df)
    # extract rows, and if needed apply post_limit
    row_iter = results_df.itertuples(index=False, name=None)
    if ext_proc["post_limit"]:
        row_iter = slicer(row_iter, slice(0, ext_proc["post_limit"]))
    rows = list(map(DFIterator.clean, row_iter))
    return cols, rows


def _translate_s2p_exception(exc, sql: str):
    """
    Convert errors into code=sql-error where this makes sense.
    """
    if isinstance(exc, SyntaxError):
        return SqlException(f"internal SQL compilation error: {str(exc)}", sql=sql)
    # pandas raises this one
    if exc.__class__.__name__ == "UndefinedVariableError":
        return SqlException_SqlProblem("Undefined column: %s" % str(exc), sql=sql)
    # column not found
    if isinstance(exc, KeyError) and "] are in the [columns]" in str(exc):
        col_info = re.sub(r"None of \[(.+)\] are in the \[columns\]", r'\1', str(exc))
        cols = []
        for col in re.finditer(r'\[([^\]]+)\]', col_info):
            col_list = col.group(1)
            cols += list(map(lambda c: c.strip(" \t'"), col_list.split(",")))
        return SqlException_SqlProblem("Undefined column(s): %s" % (", ".join(cols)), sql=sql)
    if isinstance(exc, KeyError):
        return SqlException_SqlProblem("Unrecognized column: %s" % exc.args[0], sql=sql)
    return exc


def read_sheet_names(input_stream):
    """
    Read sheet names from an XLSX file.
    :returns:  A tuple of column_names, row_values
    """
    book = openpyxl.open(input_stream)
    return ["name"], [[name] for name in book.sheetnames]


def total_table_stream_sizes(open_table_stream: callable, sql: str, external_data_indicator: int=1000000000):
    """
    Determine total size of all referenced tables.
    """
    total = 0
    for table_spec in find_referenced_tables(sql):
        with open_table_stream(table_spec) as stream:
            if isinstance(stream, ChunkedRowIterator):
                return external_data_indicator
            stream.seek(0, 2)
            total += stream.tell()
    return total



"""
Utilities for evaluation of SQL.
"""
from dslibrary.sql.data import SqlDataSupplier
from dslibrary.sql.exceptions import SqlException_DataSourceNotFound

# operators
from dslibrary.sql.misc import cmp_table_names

OPS = [
    ["or"],
    ["and"],
    ["=", "==", "!=", "<>", "is", "is not", "in", "not in", "like", "not like", "ilike", "not ilike", "glob", "not glob", "match",
     "not match", "regexp", "not regexp"],
    ["<", "<=", ">", ">="],
    ["<<", ">>", "&", "|"],
    ["+", "-"],
    ["*", "/", "%"],
    ["||"]
]
OPS_ALL = set([op for ops in OPS for op in ops]) | {"between"}
OP_PRECEDENCE = {op: n for n, opGroup in enumerate(OPS) for op in opGroup}
UOPS = {"-", "+", "~", "not"}


class SqlEvaluationContext(object):
    class NullCache(object):
        def __getitem__(self, item):
            """ All requests return None """

        def __setitem__(self, key, value):
            """ Storing item has no effect """

        def get(self, item):
            """ Cache is always empty """

        def __contains__(self, item):
            return False

    def __init__(self, parent=None, record=None, fields=None, table_aliases=None, compiler=None, data_suppliers=None,
                 aggregate=None, debug_level=None, ext_vars=None, cache=None):
        self.parent = parent
        self._record = record
        self._fields = fields
        self._table_aliases = table_aliases
        self._compiler = compiler
        self._data_suppliers = data_suppliers
        self._cache = cache
        self._debug_level = debug_level
        self._aggregate = aggregate
        self._ext_vars = ext_vars

    def getRoot(self):
        """
        Finds the topmost context in a tree.
        """
        if self.parent:
            return self.parent.getRoot()
        return self

    def _get_debug_level(self):
        return self.getRoot()._debug_level or 0

    def _set_debug_level(self, v):
        self.getRoot()._debug_level = v

    debugLevel = property(_get_debug_level, _set_debug_level)
    """
    Determines whether to display diagnostic information.  This value
    is managed only at the root level.
    """

    @property
    def cache(self):
        """
        Information about the current operation can be cached here.  This accessor returns the
        cache from the topmost context object, i.e. a global cache for the whole operation.
        """
        obj = self.getRoot()
        if obj._cache is None:
            obj._cache = {}
        return obj._cache

    @property
    def fieldCache(self):
        """
        Returns the cache for the topmost context which defines a list of fields.
        """
        if self._fields is not None:
            if self._cache is None:
                self._cache = {}
            return self._cache
        if self.parent:
            return self.parent.fieldCache

    def _get_record(self):
        if self._record:
            return self._record
        if self.parent:
            return self.parent.record

    def _set_record(self, v):
        self._record = v

    record = property(_get_record, _set_record)
    """
    The current data row.
    """

    def _get_aggregate(self):
        if self._aggregate is not None:
            return self._aggregate
        if self.parent:
            return self.parent.aggregate
        return True

    def _set_aggregate(self, v):
        self._aggregate = v

    aggregate = property(_get_aggregate, _set_aggregate)
    """
    Whether to apply values to aggregators, or to only return current values.
    """

    def _get_compiler(self):
        if self._compiler:
            return self._compiler
        if self.parent:
            return self.parent.compiler

    def _set_compiler(self, v):
        self._compiler = v

    compiler = property(_get_compiler, _set_compiler)
    """
    A compiler which can be used to create data manipulation structures on the fly.
    """

    def _get_data_suppliers(self):
        if self._data_suppliers is not None:
            return self._data_suppliers
        if self.parent:
            return self.parent.data_suppliers
        else:
            self._data_suppliers = []
            return self._data_suppliers

    def _set_data_suppliers(self, v):
        self._data_suppliers = v

    data_suppliers = property(_get_data_suppliers, _set_data_suppliers)
    """
    The list of data suppliers which stream data from available sources.
    """

    def _get_fields(self):
        if self._fields:
            return self._fields
        if self.parent:
            return self.parent.fields

    def _set_fields(self, v):
        self._fields = v

    fields = property(_get_fields, _set_fields)
    """
    Definitions for each field in the current row or stream.
    """

    def _get_table_aliases(self):
        if self._table_aliases is not None:
            return self._table_aliases
        if self.parent:
            return self.parent.table_aliases
        self._table_aliases = {}
        return self._table_aliases

    def _set_table_aliases(self, v):
        self._table_aliases = v

    table_aliases = property(_get_table_aliases, _set_table_aliases)
    """
    A set of aliases that apply to all known tables.  This is a mapping from aliases to full table names.
    """

    def getNamedValue(self, fieldSpec):
        """
        Look for the given value, which is in the form of a tuple of tableName/tableAlias and fieldName/fieldAlias.

        If the value is not found in the current record, we check the parent context, which may contain a different
        record, in the case where we are a subquery context.

        :param fieldSpec:  A tuple indicating a table and field name.
        """
        idx = find_field_in_list(fieldSpec, self)
        if idx >= 0:
            if self.record and idx < len(self.record):
                return self.record[idx]
        elif self.parent:
            return self.parent.getNamedValue(fieldSpec)

    def namedValueAvailable(self, fieldSpec):
        """
        Look for the given value, which is in the form of a tuple of tableName/tableAlias and fieldName/fieldAlias.

        If the value is not found in the current record, we check the parent context, which may contain a different
        record, in the case where we are a subquery context.
        """
        idx = find_field_in_list(fieldSpec, self)
        if idx >= 0:
            return True
        elif self.parent:
            return self.parent.namedValueAvailable(fieldSpec)
        return False

    def getExtVar(self, name: str):
        """
        Get the value of a named external variable.
        """
        if not self._ext_vars or name not in self._ext_vars:
            raise Exception("variable not found: %s" % name)
        return self._ext_vars.get(name)

    def derive(self, record=None, fields=None, tableAliases=None, aggregate=None, cache=None):
        """
        Create a new derived context which inherits all properties from 'self'.
        """
        return SqlEvaluationContext(parent=self, record=record, fields=fields, table_aliases=tableAliases,
                                    aggregate=aggregate, ext_vars=self._ext_vars, cache=cache)

    def addDataSupplier(self, supplier):
        """
        All data ultimately comes from the suppliers added here.
        """
        assert isinstance(supplier, SqlDataSupplier)
        self.data_suppliers.append(supplier)

    def findDataSource(self, tableSpec, errorIfNotFound=True):
        """
        Find a data source from one of the registered data suppliers.
        """
        for supplier in self.data_suppliers:
            found = supplier.getTableStream(tableSpec)
            if found:
                return found
        if errorIfNotFound:
            raise SqlException_DataSourceNotFound(
                "data source not found: {0}".format(".".join(filter(lambda x: x, tableSpec)) if tableSpec else "None"))


def find_field_in_list(field_spec, context):
    """
    Find the index of a matching field.
    :param field_spec:   A tuple with at least a field name and optionally a database and table name.
    :param context:     A context which contains a field list to scan.
    :return:  -1 if not found, or index in field list.
    """
    if isinstance(field_spec, str):
        field_spec = (field_spec,)
    if not context or not context.fields:
        return -1

    field_cache = context.fieldCache
    if field_cache is not None:
        out = field_cache.get(field_spec)
        if out is not None:
            return out
        out = _find_field_in_list(field_spec, context)
        field_cache[field_spec] = out
        return out
    else:
        return _find_field_in_list(field_spec, context)


def _find_field_in_list(field_spec, context):
    table_spec = field_spec[:-1] if len(field_spec) > 1 else None
    field_name = field_spec[-1]
    # look up table name from alias
    derived_alias = None
    if table_spec and context.table_aliases:
        derived_alias = context.table_aliases.get(table_spec[-1])
        if isinstance(derived_alias, (tuple, list)):
            derived_alias = derived_alias[-1]
    # scan
    n_best, v_best = -1, 0
    # match field alias first
    for n, field in enumerate(context.fields):
        w = match_spec_to_field(table_spec, field_name, field, derived_alias)
        if w > v_best:
            n_best, v_best = n, w
    return n_best


def match_spec_to_field(table_spec, field_spec, field, derived_alias=None):
    """
    Compare a requested field with an available field.
    :param table_spec:  A tuple giving the components of a table, i.e. (database,table) or (table,).  The table name
            may also be the alias for a table.
    :param field_spec:  A string with the name of a field or an alias.
    :param field:      The available field to which we are comparing.
    :param derived_alias:  The implied alias of the supplied table name.  This alias may be from a different context so
            it is to be weighed lightly.
    :return:  A weighting, with higher numbers indicating better matches, and 0 indicating no match.
    """
    # compare field name - at least a loose match is mandatory
    w_f = _field_match_ranking(field_spec, field)
    if not w_f:
        return 0
    # check database name
    w_d = _database_match_ranking(table_spec, field)
    if w_d < 0:
        return 0
    # compare table name
    if not table_spec or not table_spec[-1]:
        # no table spec at all - this is the weakest match
        w_t = 0
    else:
        w_t = _table_match_ranking(table_spec, field, derived_alias)
        if w_t == 0:
            return 0
        if w_t == 3 and w_d == 0:
            # if there's a table alias we don't need the database to be specified
            w_d = 1
    # add the points together
    return w_f + w_t + w_d


def _field_match_ranking(field_spec, field):
    # compare field name
    if field.alias and cmp_table_names(field_spec, field.alias):
        # match on alias is best
        return 3
    elif cmp_table_names(field_spec, field.field):
        if not field.alias:
            # match on field name is 2nd best
            return 2
        else:
            # if there's an alias but we match the field name that's 'acceptable' but the weakest possible match
            return 1
    else:
        return 0


def _table_match_ranking(table_spec, field, derived_alias):
    if field.table_alias and cmp_table_names(table_spec[-1], field.table_alias):
        # match on table alias is the best
        return 3
    elif cmp_table_names(table_spec[-1], field.table):
        if not field.table_alias:
            # 2nd best is match on table name when there is no table alias
            return 2
        else:
            # 3rd is match on table name when we could have matched an alias
            return 1
    else:
        if cmp_table_names(derived_alias, field.table_alias):
            # 4th - indirectly implied table alias matches
            return 1
        else:
            # mismatch on table name: reject
            return 0


def _database_match_ranking(table_spec, field):
    db_name = table_spec[-2] if table_spec and len(table_spec) > 1 else None
    if field.database or db_name:
        if field.database and db_name:
            # if a database name is specified on both sides it must match
            if not cmp_table_names(db_name, field.database):
                return -1
            else:
                # proper match on database
                return 2
        else:
            # a database name is specified on only one side - this is not great but we can work around it
            return 0
    else:
        # no database specified
        return 1



import time
import re
import math
import unicodedata
from dateutil.parser import parse as parse_date

from .misc import sql_enquote_id, enquote_string_literal, parse_iso_datetime
from .eval_utils import OP_PRECEDENCE, SqlEvaluationContext


ALL_FUNCTIONS = frozenset({
    # GENERAL/LOGICAL
    "typeof", "not", "coalesce", "isnull", "ifnull", "nullif", "if",
    # STRING
    "length", "len",
    "lcase", "lower", "ucase", "upper", "left", "right",
    "concat", "hex",
    "glob", "zeroblob", "randomblob",
    "trim", "char", "chr", "instr", "like", "ltrim", "quote", "replace", "rtrim",
    "regexp_replace", "soundex", "substr", "unicode",
    # MATH
    "abs", "sqrt", "least", "greatest", "round", "floor", "ceil", "ceiling", "random",
    "mod", "pow", "pi", "sin", "cos", "tan", "asin", "acos", "atan", "exp", "log",
    # DATE/TIME
    "strftime", "date", "time", "datetime", "julianday", "now", "year", "month", "day"
})
FUNCTION_ALIASES = {
    "len": "length",
    "ifnull": "isnull",
    "coalesce": "isnull",
    "chr": "char",
    "lcase": "lower",
    "ucase": "upper",
    "ceiling": "ceil",
}
OPERATORS = {
    "||": lambda left, right: str(left) + str(right),
    "=": lambda left, right: left == right,
    "==": lambda left, right: left == right,
    "<": lambda left, right: left < right,
    ">": lambda left, right: left > right,
    "<=": lambda left, right: left <= right,
    ">=": lambda left, right: left >= right,
    "!=": lambda left, right: left != right,
    "<>": lambda left, right: left != right,
    "and": lambda left, right: left and right,
    "or": lambda left, right: left or right,
    "+": lambda left, right: str(left) + str(right) if isinstance(left, str) or isinstance(right, str) else left + right,
    "-": lambda left, right: left - right,
    "*": lambda left, right: left * right,
    "/": lambda left, right: left / right if right else None,
    "%": lambda left, right: left % right if right else None,
    "<<": lambda left, right: left << right,
    ">>": lambda left, right: left >> right,
    "&": lambda left, right: left & right,
    "|": lambda left, right: left | right,
    "in": lambda left, right: left in right,
    "not in": lambda left, right: left not in right,
    "like": lambda left, right: _like(left, right),
    "not like": lambda left, right: not _like(left, right),
    "ilike": lambda left, right: _ilike(left, right),
    "not ilike": lambda left, right: not _ilike(left, right),
    "regexp": lambda left, right: re.search(str(right), str(left)) is not None,
    "not regexp": lambda left, right: re.search(str(right), str(left)) is None,
    "match": lambda left, right: re.match("^" + str(right) + "$", str(left)) is not None,
    "not match": lambda left, right: re.match("^" + str(right) + "$", str(left)) is None
}


class EvalContext(object):
    def __init__(self):
        self.current_row = tuple()
        self.current_fields = tuple()


class Node(object):
    def eval(self, ctx):
        return None

    def _cloneOfChildren(self):
        if hasattr(self, "children"):
            return [c.clone() for c in self.children]

    def simplify(self, context):
        if hasattr(self, "children"):
            vals = []
            self.children = [sub.simplify(context) for sub in self.children]
            for sub in self.children:
                if not isinstance(sub, Node_Literal):
                    return self
                vals.append(sub.value)
            return Node_Literal(self.eval(context))
        return self

    def to_sql(self, flavor: str=None):
        return ""

    def __str__(self):
        return self.to_sql()


class Node_Literal(Node):
    def __init__(self, value):
        self.value = value

    def eval(self, ctx):
        return self.value

    def clone(self):
        return Node_Literal(self.value)

    def to_sql(self, flavor: str=None):
        if self.value is None:
            return "NULL"
        if self.value is False:
            return "FALSE"
        if self.value is True:
            return "TRUE"
        if isinstance(self.value, str):
            return enquote_string_literal(self.value)
        else:
            return str(self.value)


class Node_Placeholder(Node):
    def __init__(self):
        """
        A placeholder for a literal value.
        """

    def eval(self, ctx):
        return None

    def clone(self):
        return Node_Placeholder()

    def to_sql(self, flavor: str=None):
        # TODO some day we will have a SQL flavor that uses ? instead of %s
        return "%s"


class Node_Field(Node):
    def __init__(self, field_name, quotes=None):
        """
        A reference to a field.
        :param field_name:  Name of field, as a tuple of strings.  Last one is field name.  Others qualify with
                database, schema, etc..
        :param quotes:      Optional iterable which indicates whether quotes were used for each name component.
        """
        if isinstance(field_name, str):
            field_name = (field_name,)
        self.fieldName = field_name
        self.quotes = quotes

    def isWildcard(self):
        return self.fieldName[-1].startswith("*")

    def getWildcardExceptions(self):
        if self.isWildcard():
            spec = self.fieldName[-1][1:]
            if spec.startswith("["):
                spec = spec[1:]
                if spec.endswith("]"):
                    spec = spec[:-1]
                return spec.split(',')

    def eval(self, ctx):
        if self.isWildcard():
            return 0
        if not ctx:
            raise ValueError(f"Cannot evaluate field='{self.fieldName}' in a literal context")
        return ctx.getNamedValue(self.fieldName)

    def clone(self):
        return Node_Field(self.fieldName, self.quotes)

    def to_sql(self, flavor: str=None):
        quotes = self.quotes or (False,) * len(self.fieldName)
        return ".".join(sql_enquote_id(part, flavor, always) for always, part in zip(quotes, self.fieldName) if part)


class Node_ExtVar(Node):
    """ Reference to an external variable """
    def __init__(self, var_name):
        self.var_name = var_name

    def eval(self, ctx):
        return ctx.getExtVar(self.var_name)

    def clone(self):
        return Node_ExtVar(self.var_name)

    def simplify(self, context):
        return Node_Literal(self.eval(context))

    def to_sql(self, flavor: str=None):
        return "@%s" % sql_enquote_id(self.var_name)


class Functions(object):
    def __init__(self, context):
        self.ctx = context

    def call(self, fn_name, arguments):
        method = "f_%s" % FUNCTION_ALIASES.get(fn_name, fn_name)
        if not hasattr(self, method):
            raise Exception("function not implemented: " + fn_name)
        fn = getattr(self, method)
        args = [arg.eval(self.ctx) for arg in arguments] if arguments else []
        return fn(*args)

    def f_typeof(self, v):
        if v is None:
            return "null"
        if isinstance(v, int):
            return "integer"
        if isinstance(v, float):
            return "real"
        if isinstance(v, str):
            return "text"
        raise Exception("unknown sql type: " + repr(type(v)))

    def f_lower(self, v):
        return str(v).lower()

    def f_upper(self, v):
        return str(v).upper()

    def f_trim(self, v):
        return str(v).strip()

    def f_char(self, v):
        return chr(int(v))

    def f_concat(self, *args):
        return "".join(map(str, args))

    def f_left(self, v1, v2):
        if not v2 or v2 < 0:
            return ""
        return str(v1)[:int(v2)]

    def f_right(self, v1, v2):
        if not v2 or v2 < 0:
            return ""
        return str(v1)[-int(v2):]

    def f_replace(self, v1, v2, v3):
        return str(v1).replace(str(v2), str(v3))

    def f_substr(self, v, start, count=None):
        if start > 0:
            start -= 1
        v = str(v)[start:]
        if count is not None and count < len(v):
            v = v[:count]
        return v

    def f_rtrim(self, v):
        return str(v).rstrip()

    def f_ltrim(self, v):
        return str(v).lstrip()

    def f_instr(self, v1, v2):
        return str(v1).find(v2) + 1

    def f_soundex(self, v):
        return soundex(v)

    def f_not(self, v):
        return not v

    def f_length(self, v):
        return len(str(v))

    def f_abs(self, v):
        return abs(v)

    def f_mod(self, v1, v2):
        if v1 is None:
            return
        if not v2:
            return
        return v1 % v2

    def f_pow(self, v1, v2):
        if v1 is None or v2 is None:
            return
        return v1 ** v2

    def f_pi(self):
        return math.pi

    def f_sin(self, v1):
        return math.sin(v1)

    def f_cos(self, v1):
        return math.cos(v1)

    def f_tan(self, v1):
        return math.tan(v1)

    def f_asin(self, v1):
        return math.asin(v1)

    def f_acos(self, v1):
        return math.acos(v1)

    def f_atan(self, v1):
        return math.atan(v1)

    def f_exp(self, v1):
        return math.exp(v1)

    def f_log(self, v1):
        return math.log(v1)

    def f_if(self, *args):
        cond = args[0]
        t = args[1] if len(args) >= 2 else 1
        f = args[2] if len(args) >= 3 else None
        return t if cond else f

    def f_nullif(self, *args):
        if len(args) < 2:
            return args[0]
        if args[0] == args[1]:
            return None
        return args[0]

    def f_isnull(self, *args):
        v1 = args[0]
        if len(args) < 2:
            return v1
        v2 = args[1]
        return v2 if v1 is None else v1

    def f_sqrt(self, v):
        return math.sqrt(v)

    def f_round(self, v, digits=None):
        if digits is not None:
            return round(v, digits)
        return round(v)

    def f_floor(self, v):
        return math.floor(v)

    def f_ceil(self, v):
        return math.ceil(v)

    def f_least(self, *args):
        return min(args)

    def f_greatest(self, *args):
        return max(args)

    def f_strftime(self, fmt, ts):
        t = _parse_time(ts)
        return time.strftime(fmt, t)

    def f_year(self, ts):
        return _parse_time(ts).tm_year

    def f_month(self, ts):
        return _parse_time(ts).tm_mon

    def f_day(self, ts):
        return _parse_time(ts).tm_mday

    def f_date(self, v1):
        t = _parse_time(v1)
        return time.strftime('%Y-%m-%d', t)

    def f_time(self, v1):
        t = _parse_time(v1)
        return time.strftime('%H:%M:%S', t)

    def f_datetime(self, v1):
        t = _parse_time(v1)
        return time.strftime('%Y-%m-%d %H:%M:%S', t)

    def f_now(self):
        t = time.gmtime(time.time())
        return time.strftime('%Y-%m-%d %H:%M:%SZ', t)

    def f_julianday(self, v1):
        t = _parse_time(v1)
        return time.mktime(t) / 86400 + 2440587.5

    def f_regexp_replace(self, expr, pattern, replacement):
        return re.sub(pattern, replacement, expr)


class Node_Function(Node):
    def __init__(self, function_name, children=None):
        self.functionName = function_name
        agg = AGGREGATORS.get(function_name)
        if agg:
            self.aggregator = agg()
        else:
            self.aggregator = None
        self.children = children or []

    def isAggregate(self):
        return self.aggregator is not None

    def eval(self, ctx):
        if self.aggregator:
            if not ctx or ctx.aggregate:
                v1 = self.children[0].eval(ctx) if self.children else 0
                self.aggregator.applyValue(v1)
                return 0
            else:
                out = self.aggregator.getResult()
                self.aggregator.reset()
                return out
        '''
           UNSUPPORTED SQLite functions:

           "char", "glob", "hex", "like",
           "quote", "random", "randomblob",
           "unicode", "zeroblob"
        '''
        return Functions(ctx).call(self.functionName, self.children)

    def clone(self):
        return Node_Function(self.functionName, self._cloneOfChildren())

    def to_sql(self, flavor: str=None):
        if self.functionName == "count_distinct":
            return "count(distinct %s)" % ", ".join(sub.to_sql(flavor) for sub in self.children)
        else:
            return "%s(%s)" % (self.functionName, ", ".join(sub.to_sql(flavor) for sub in self.children))


class Node_Unary(Node):
    def __init__(self, child, operator):
        self.children = [child]
        self.operator = operator

    def eval(self, ctx):
        sub = self.children[0].eval(ctx)
        if sub is None:
            return None
        if self.operator == "-":
            if isinstance(sub, (int, float)):
                return -sub
            else:
                raise Exception("'-' applied to non-numeric value")
        if self.operator == "+":
            if isinstance(sub, (int, float)):
                return sub
            else:
                raise Exception("'+' applied to non-numeric value")
        if self.operator == "~":
            if isinstance(sub, (int, float)):
                return ~int(sub)
            else:
                raise Exception("'~' applied to non-numeric value")
        if self.operator == "not":
            return not sub
        raise Exception("operator not implemented: " + self.operator)

    def clone(self):
        return Node_Unary(self.children[0].clone(), self.operator)

    def to_sql(self, flavor: str=None):
        return self.operator + ' ' + self.children[0].to_sql(flavor)


class Node_Cast(Node):
    def __init__(self, child, to_type):
        self.children = [child]
        self.to_type = (to_type or "").lower()

    def eval(self, ctx):
        sub = self.children[0].eval(ctx)
        base_type = self.to_type.split("(")[0]
        if sub is None:
            return None
        if base_type in {"int", "tinyint", "smallint", "bigint", "integer"}:
            try:
                return int(sub)
            except:
                raise Exception("cannot cast %s to int" % type(sub))
        if base_type in {"float", "double", "real"}:
            try:
                return float(sub)
            except:
                raise Exception("cannot cast %s to float" % type(sub))
        if base_type in {"string", "str", "varchar"} or self.to_type.startswith("varchar"):
            try:
                return str(sub)
            except:
                raise Exception("cannot cast %s to str" % type(sub))
        if base_type in {"datetime", "timestamp"}:
            try:
                return parse_iso_datetime(str(sub))
            except:
                raise Exception("cannot cast %s to str" % type(sub))
        raise Exception("type not recognized: " + self.to_type)

    def clone(self):
        return Node_Cast(self.children[0].clone(), self.to_type)

    def to_sql(self, flavor: str=None):
        return "CAST(%s as %s)" % (self.children[0].to_sql(flavor), self.to_type.upper())


class Node_Select(Node):
    def __init__(self, select_spec):
        self.select_spec = select_spec
        self.id = str(select_spec)
        self.cacheable = None

    def getValueList(self, ctx):
        if self.cacheable and self.id in ctx.cache:
            select_result = ctx.cache[self.id]
        else:
            strm = ctx.compiler.compile_select(self.select_spec, ctx.derive(tableAliases={}, aggregate=True),
                                               subquery=True)
            if self.cacheable is None:
                # if any fields are required outside the immediate scope, this implies (or could imply) a query with moving parts
                ctx_local = SqlEvaluationContext(fields=strm.getFields(),
                                                 table_aliases=self.select_spec.getAliasToTableMapping())
                self.cacheable = self.select_spec.ok_to_cache_as_subquery(ctx_local)
            select_result = [row[0] for row in strm.getRowIterator()]
            if self.cacheable:
                ctx.cache[self.id] = select_result
        return select_result

    def eval(self, ctx):
        v = self.getValueList(ctx)
        if not v:
            return None
        if len(v) == 1:
            return v[0]
        return v

    def clone(self):
        return Node_Select(self.select_spec)

    def simplify(self, context):
        self.select_spec.simplify(context)
        return self

    def to_sql(self, flavor: str=None):
        return "({0})".format(self.select_spec.to_sql(flavor))


class Node_InSelect(Node_Select):
    def __init__(self, left, select_spec):
        Node_Select.__init__(self, select_spec)
        self.children = [left]

    def eval(self, ctx):
        select_result = self.getValueList(ctx)
        v_l = self.children[0].eval(ctx)
        return v_l in select_result

    def clone(self):
        return Node_InSelect(self.children[0].clone(), self.select_spec)

    def to_sql(self, flavor: str=None):
        return "{0} in ({1})".format(self.children[0].to_sql(flavor), self.select_spec.to_sql(flavor))


class Node_List(Node):
    def __init__(self, children=None):
        self.children = children or []

    def eval(self, ctx):
        return [v.eval(ctx) for v in self.children]

    def clone(self):
        return Node_List([c.clone() for c in self.children])

    def to_sql(self, flavor: str=None):
        return "({0})".format(", ".join([c.to_sql(flavor) for c in self.children]))


class Node_Op(Node):
    def __init__(self, operator, add_children=None):
        self.operator = operator
        self.children = add_children or []
        self.null_if_any_null = True
        f = OPERATORS.get(self.operator)
        if self.operator == "or":
            self.null_if_any_null = False
        if not f:
            def unimpl(l, r):
                raise Exception("operator not implemented yet: " + self.operator)
            f = unimpl
        self.f = f

    def eval(self, ctx):
        left = self.children[0].eval(ctx)
        right = self.children[1].eval(ctx)
        if self.null_if_any_null and (left is None or right is None):
            return None
        return self.f(left, right)

    def clone(self):
        return Node_Op(self.operator, self._cloneOfChildren())

    def to_sql(self, flavor: str=None):
        prec = OP_PRECEDENCE.get(self.operator, -1)
        def child(n):
            parens = False
            ch = self.children[n]
            if isinstance(ch, Node_Op):
                c_prec = OP_PRECEDENCE.get(ch.operator, -1)
                if c_prec < prec:
                    # if child operator has lower precedence it needs parentheses
                    parens = True
            if parens:
                return f"({ch.to_sql(flavor)})"
            else:
                return ch.to_sql(flavor)
        return child(0) + ' ' + self.operator + ' ' + child(1)


class Node_Op_Is(Node):
    OP_IS = "is"
    OP_IS_NOT = "is not"
    def __init__(self, operator, add_children=None):
        self.operator = operator
        self.children = add_children or []
        if self.operator == self.OP_IS:
            self.f = lambda left, right: left == right
        elif self.operator == self.OP_IS_NOT:
            self.f = lambda left, right: left != right

    def eval(self, ctx):
        left = self.children[0].eval(ctx)
        right = self.children[1].eval(ctx)
        return self.f(left, right)

    def clone(self):
        return Node_Op_Is(self.operator, self._cloneOfChildren())

    def to_sql(self, flavor: str=None):
        return self.children[0].to_sql(flavor) + ' ' + self.operator + ' ' + self.children[1].to_sql(flavor)


class Node_Case(Node):
    def __init__(self, when_then_else):
        self.children = when_then_else or []

    def eval(self, ctx):
        for idx in range(0, len(self.children) - 1, 2):
            if self.children[idx].eval(ctx):
                return self.children[idx + 1].eval(ctx)
        if len(self.children) % 2 == 1:
            return self.children[-1].eval(ctx)

    def clone(self):
        return Node_Case(self._cloneOfChildren())

    def to_sql(self, flavor: str=None):
        parts = ["CASE"]
        for idx in range(0, len(self.children) - 1, 2):
            parts.append("WHEN {0} THEN {1}".format(self.children[idx].to_sql(flavor), self.children[idx + 1].to_sql(flavor)))
        if len(self.children) % 2 == 1:
            parts.append("ELSE {0}".format(self.children[-1].to_sql(flavor)))
        parts.append("END")
        return " ".join(parts)


class SqlOrderBy(object):
    def __init__(self):
        self.levels = []

    def addLevel(self, expr, ascending=True):
        self.levels.append((expr, ascending))

    def compare(self, rec_l, rec_r, ctx):
        for level in self.levels:
            v_l = level[0].eval(ctx.derive(record=rec_l))
            v_r = level[0].eval(ctx.derive(record=rec_r))
            c = _cmp(v_l, v_r)
            if not level[1]:
                c = -c
            if c != 0:
                return c
        return 0

    def to_sql(self, flavor: str=None):
        out = []
        for level in self.levels:
            part = level[0].to_sql(flavor)
            if not level[1]:
                part += " DESC"
            out.append(part)
        return ", ".join(out)


def _parse_time(timestr, ignoretz=True):
    if timestr == "now":
        return time.time()
    dt = parse_date(timestr, ignoretz=ignoretz)
    return dt.timetuple()


def _like(value, ptn):
    ptn = str(ptn)
    ptn = re.sub(r'([.+\-*?\[\]()|{\}~^$])', r'\\\1', ptn)
    ptn = ptn.replace("%", ".*").replace("_", ".")
    return re.match(ptn, str(value)) is not None


def _ilike(value, ptn):
    ptn = str(ptn)
    ptn = re.sub(r'([.+\-*?\[\]()|{\}~^$])', r'\\\1', ptn)
    ptn = ptn.replace("%", ".*").replace("_", ".")
    return re.match(ptn, str(value), re.IGNORECASE) is not None


class SqlAggregator(object):
    """
    Base class for aggregation methods.
    """
    def reset(self):
        """
        Reset state for another aggregation run.
        """

    def applyValue(self, value):
        """
        Aggregate one more value.
        """

    def getResult(self):
        """
        Get final aggregated result.
        """


class _AggMin(SqlAggregator):
    def __init__(self):
        self.reset()

    def reset(self):
        self.value = None

    def applyValue(self, value):
        if value is not None and (self.value is None or value < self.value):
            self.value = value

    def getResult(self):
        return self.value


class _AggMax(SqlAggregator):
    def __init__(self):
        self.reset()

    def reset(self):
        self.value = None

    def applyValue(self, value):
        if value is not None and (self.value is None or value > self.value):
            self.value = value

    def getResult(self):
        return self.value


class _AggCount(SqlAggregator):
    def __init__(self):
        self.reset()

    def reset(self):
        self.value = 0

    def applyValue(self, value):
        if value is not None:
            self.value += 1

    def getResult(self):
        return self.value


class _AggCountDistinct(SqlAggregator):
    def __init__(self):
        self.reset()

    def reset(self):
        self.values = set()

    def applyValue(self, value):
        if value is not None:
            if isinstance(value, list):
                value = tuple(value)
            self.values.add(value)

    def getResult(self):
        return len(self.values)


class _AggSum(SqlAggregator):
    def __init__(self):
        self.reset()

    def reset(self):
        self.value = 0

    def applyValue(self, value):
        if isinstance(value, (int, float)):
            self.value += value

    def getResult(self):
        return self.value


class _AggAvg(SqlAggregator):
    def __init__(self):
        self.reset()

    def reset(self):
        self.value = 0
        self.count = 0

    def applyValue(self, value):
        if isinstance(value, (int, float)):
            self.value += value
            self.count += 1

    def getResult(self):
        if self.count:
            return float(self.value) / self.count


class _AggMedian(SqlAggregator):
    def __init__(self):
        self.reset()

    def reset(self):
        self.values = []

    def applyValue(self, value):
        if value is not None:
            self.values.append(value)

    def getResult(self):
        self.values.sort()
        n = len(self.values)
        if not n:
            return None
        if n == 1:
            return self.values[0]
        if n % 2 == 1:
            return self.values[n//2]
        v0 = self.values[n//2-1]
        v1 = self.values[n//2]
        if isinstance(v0, str):
            return v0 + " "
        return (v0+v1)/2


class _AggStdev(SqlAggregator):
    def __init__(self):
        self.reset()

    def reset(self):
        self.mean = 0.0
        self.squares = 0.0
        self.n = 1

    def applyValue(self, value):
        if isinstance(value, (int, float)):
            t_m = self.mean
            self.mean += (value - t_m) / self.n
            self.squares += (value - t_m) * (value - self.mean)
            self.n += 1

    def getResult(self):
        if self.n < 3:
            return None
        return math.sqrt(self.squares / (self.n - 2))


AGGREGATORS = {
    "min": _AggMin,
    "max": _AggMax,
    "count": _AggCount,
    "count_distinct": _AggCountDistinct,
    "sum": _AggSum,
    "total": _AggSum,
    "avg": _AggAvg,
    "average": _AggAvg,
    "median": _AggMedian,
    "stdev": _AggStdev
}
AGGREGATE_FNS = frozenset(AGGREGATORS.keys())


def _is_literal(node):
    if isinstance(node, Node_Literal):
        return True
    if hasattr(node, "children"):
        for sub in node.children:
            if not _is_literal(sub):
                return False
        return True
    return False


def simplify_expression(node, context):
    """
    Simplify an expression.  Currently only collapses operations between literals.
    """
    def simplify(node):
        if not isinstance(node, Node_Literal) and _is_literal(node):
            value = node.eval(context)
            return Node_Literal(value)
        if hasattr(node, "children"):
            node.children = [simplify(sub) for sub in node.children]
        return node
    return simplify(node)


def has_any_aggregators(node):
    """
    Detect the presence of aggregating functions.
    """
    if isinstance(node, Node_Function) and node.aggregator:
        return True
    if hasattr(node, "children"):
        for sub in node.children:
            if has_any_aggregators(sub):
                return True
    return False


def uses_aggregation(select_spec):
    """
    Check whether a SELECT uses any aggregation.
    """
    # if grouping is requested, then we don't need to look further
    if select_spec.group_by:
        return True
    # check for aggregation functions
    if not select_spec.fields:
        return False
    for f in select_spec.fields:
        if has_any_aggregators(f[0]):
            return True
    return False


def _cmp(a, b):
    if a is None and b is None:
        return 0
    if b is None:
        return -1
    if a is None:
        return 1
    try:
        return (a > b) - (a < b)
    except TypeError:
        a, b = str(a), str(b)
        return (a > b) - (a < b)


def soundex(name, len=4):
    """ soundex module conforming to Knuth's algorithm
        implementation 2000-12-24 by Gregory Jorgensen
        public domain

        source: http://code.activestate.com/recipes/52213-soundex-algorithm/
    """
    #
    name = strip_accents(name)
    # digits holds the soundex values for the alphabet
    digits = '01230120022455012623010202'
    sndx = ''
    fc = ''
    # translate alpha chars in name to soundex digits
    for c in name.upper():
        if c.isalpha():
            if not fc: fc = c  # remember first letter
            d = digits[ord(c) - ord('A')]
            # duplicate consecutive soundex digits are skipped
            if not sndx or (d != sndx[-1]):
                sndx += d
    # replace first digit with first alpha character
    sndx = fc + sndx[1:]
    # remove all 0s from the soundex code
    sndx = sndx.replace('0', '')
    # return soundex code padded to len characters
    return (sndx + (len * '0'))[:len]


### copied from MiscUtils, to make sql package standalone ###
PTN_ANYUNICODE = re.compile(r'[\x80-\xFF]')
PTN_ANYUNICODE_u = re.compile(r'[\u0080-\uFFFF]')


def strip_accents(input_str):
    if isinstance(input_str, str):
        if PTN_ANYUNICODE.search(input_str) is None:
            return input_str
    else:
        if PTN_ANYUNICODE_u.search(input_str) is None:
            return str(input_str)
    # source: http://stackoverflow.com/questions/517923/what-is-the-best-way-to-remove-accents-in-a-python-unicode-string
    nkfd_form = unicodedata.normalize('NFKD', input_str)
    s = u"".join([c for c in nkfd_form if not unicodedata.combining(c)])
    return str(s.encode('ASCII', 'ignore'))





"""
General purpose utilities with no serious dependencies.
"""
import re
import time
import datetime
import dateutil.parser
import urllib.parse

from dslibrary.utils.format_sniffer import find_url_extension

SQL_PARTS = re.compile(r'(/\*.*?\*/|"([^"]|"")*?"|`[^`]*?`|\'([^\']|\'\')*?\'|((?!/\*)(?!%s)(?!\?)[^;\'"`])+|;|%s|\?)',
                       re.DOTALL)
PTN_NOQUOTES = re.compile(r'^[a-zA-Z_][a-zA-Z0-9_]*$')
PTN_NOQUOTES_PG = re.compile(r'^[a-z_][a-z0-9_]*$')


def sql_split(operation, parameters=None):
    """
    Break apart SQL statements.  Returns a generator of SQL strings.
    """
    params_in = list(parameters or [])
    bits = []
    params = []
    for part in SQL_PARTS.finditer(operation):
        bit = part.group(0)
        if bit == ";":
            if bits:
                yield "".join(bits).strip(), params
            bits.clear()
            params = []
        else:
            if bit in {"?", "%s"}:
                params.append(params_in.pop(0) if params_in else None)
            bits.append(bit)
    if bits:
        yield "".join(bits).strip(), params


def sql_verb(sql: str):
    """
    Determine what the 'verb' is for a SQL statement (SELECT, INSERT, etc..)
    """
    for part in SQL_PARTS.finditer(sql):
        bit = part.group(0).strip()
        if bit.startswith("/*"):
            continue
        return re.split(r'\s+', bit)[0].upper()
    return ""


def is_sql_read_operation(operation):
    """
    Detect a read-only SQL operation.
    """
    valid = False
    for stmt, _ in sql_split(operation):
        verb = sql_verb(stmt)
        if verb not in ("SELECT", "SHOW", "DESCRIBE"):
            return False
        valid = True
    return valid


def sql_enquote_id(s: (str, list, tuple), mode: str=None, always: bool=False, dotted: bool=False):
    """
    Enquote an identifier (table or column name) according to different conventions.
    :param s:       What to enquote.
    :param mode:    'mysql', 'postgres', 'ansi', ...
    :param always:  True to use quotes even when not required.
    :param dotted:  True when the name may contain multiple components, i.e. database.schema.table.  Otherwise dots are
                    considered part of the name to enquote.
    :return:    Identifier, quoted if necessary.
    """
    if isinstance(s, (list, tuple)):
        return ".".join(map(lambda sn: sql_enquote_id(sn, mode=mode, always=always), s))
    if dotted:
        return ".".join(map(lambda s1: sql_enquote_id(s1, mode=mode, always=always), s.split(".")))
    if s == "*" or s.startswith("*["):
        # special cases
        return s
    if (PTN_NOQUOTES_PG if mode == "postgres" else PTN_NOQUOTES).match(s) is not None and not always:
        return s
    # NOTE: For 'mssql', the old style is to use '[]' around identifiers, which we follow.
    #  - there is a flag to have it support double quotes (the ISO standard) but one can't count on it being set
    if mode == "mssql":
        return "[" + s.replace("[", "[[").replace("]", "]]") + "]"
    # we enquote file-based data sources like MySQL for an obscure reason:
    # - pandas.read_sql() strips double-quotes out of SQL!
    q = '`' if mode in {"mysql", "s3", "abfs", "abs", "azureblob", "bigquery"} else '"'
    return "%s%s%s" % (q, s.replace(q, q+q), q)


def is_valid_field_name(field_name: str, mode: str=None):
    """
    Check whether a field name is allowed for a given target dialect.
    :param field_name:  Name of proposed field name.
    :param mode:        Dialect name -- see sql_enquote_id().
    """
    if not field_name:
        return False
    rgx = None
    if mode == "bigquery":
        # see https://cloud.google.com/bigquery/docs/schemas
        rgx = r'^[a-zA-Z_][a-zA-Z_0-9]{,127}$'
    if mode == "mysql":
        rgx = r'^[^\`\0\255\./\\]{,64}$'
    if mode == "postgres":
        rgx = r'^.{,63}$'
    if rgx and not re.match(rgx, field_name):
        return False
    return True


def is_valid_table_name(table_name: str, mode: str=None):
    """
    Check whether a table name is allowed for a given target dialect.
    """
    if not table_name:
        return False
    rgx = None
    if mode == "bigquery":
        # see https://cloud.google.com/bigquery/docs/tables
        rgx = r'^[a-zA-Z_0-9]{,1024}$'
    if mode == "mysql":
        rgx = r'^[^\`\0\255\./\\]{,64}$'
    if mode == "postgres":
        # ALSO: reserved words: Appendix B of the PostgreSQL User's Guide.
        rgx = r'^.{,63}$'
    if rgx and not re.match(rgx, table_name):
        return False
    return True


def enquote_string_literal(v):
    """
    Enquote a string for SQL.
    """
    v0 = str(v).replace("'", "''")
    p = 0
    parts = ["'"]
    for c in re.finditer(r'([\x00-\x1F\x7F-\u7fff])', v0):
        parts.append(v0[p:c.start()])
        parts.append("'||char(%d)||'" % ord(c.group(1)))
        p = c.end()
    parts.append(v0[p:])
    parts.append("'")
    out = "".join(parts)
    if out.endswith("||''"):
        out = out[:-4]
    return out


def _to_time_number(dt):
    if isinstance(dt, (int, float)):
        return dt
    elif isinstance(dt, tuple):
        return time.mktime(dt)
    elif isinstance(dt, (datetime.datetime, datetime.date)):
        return dt.timestamp()
    else:
        return


def parse_iso_datetime(v: (str, float, datetime.datetime), ignore_error: bool=False):
    """
    Parse date/time formats and produce a date number in time.time() format (seconds past 1970).
    """
    if not isinstance(v, str):
        return _to_time_number(v)
    try:
        dt = dateutil.parser.parse(v.upper(), ignoretz=True)
        return (dt - datetime.datetime(1970, 1, 1)).total_seconds()
    except Exception:
        if ignore_error:
            return
        raise


def cmp_table_names(a, b):
    """
    How to compare table and field names.  Ignores case.
    """
    if a is None or b is None:
        return False
    return a.lower() == b.lower()


def loose_table_match(value, listOrSetOrDict):
    """
    Find the closest match.
    """
    if value is None or listOrSetOrDict is None:
        return None
    # check for exact match
    if value in listOrSetOrDict:
        return value
    # scan for inexact match
    for k in listOrSetOrDict:
        if cmp_table_names(value, k):
            return k


def table_name_from_table_spec(table_spec: (str, tuple)):
    """
    Extract the name of a table from a full table specification.
    """
    if not table_spec:
        return
    return table_spec if isinstance(table_spec, str) else table_spec[-1]


def format_options_for_table(format_options: dict, table_spec: (tuple, str)):
    """
    Given a set of query-wide options, and a table specification, determine which options
    apply to the given table.  Note that the table name itself can contain options, so these
    are also extracted and applied.
    :param format_options:      A {} mapping table names to option {}s, and with the key "" representing
                                default options.
    :param table_spec:          Table specification, a tuple with parts like database, schema and table name.
    :return:        A {} with options appropriate to the indicated table.
    """
    if not table_spec:
        return
    table_name = table_spec if isinstance(table_spec, str) else table_spec[-1]
    # options can be transmitted via table name, using URI conventions
    if '#' in table_name:
        table_name, qs = table_name.split('#', maxsplit=1)
        uri_args = {k: vv[-1] if len(vv) == 1 else vv for k, vv in urllib.parse.parse_qs(qs).items()}
    else:
        uri_args = {}
    if format_options:
        if table_name in format_options:
            opts = format_options[table_name]
        else:
            opts = format_options.get("", {})
    else:
        opts = {}
    opts.update(uri_args)
    # guess file format from table name
    if "format" not in opts and "." in table_name:
        opts["format"] = find_url_extension(table_name)
    return opts


def str_to_var_name(s: str) -> str:
    """
    Convert an arbitrary string into a valid python variable name.
    """
    def xlt(m):
        n = ord(m.group(0))
        return f"_{hex(n//16)[2:]}{hex(n%16)[2:]}"
    out = re.sub(r'[^a-zA-Z0-9]', xlt, s)
    if not out or out[0].isdigit():
        out = "__" + out
    return out


def var_name_to_str(s: str) -> str:
    """
    Convert an encoded python variable name back into a string.
    """
    def xlt(seq):
        return chr(int(seq.group(0)[1:], 16))
    if s.startswith("__"):
        s = s[2:]
    return re.sub(r'_[0-9a-f][0-9a-f]', xlt, s)



import pandas

from .compiler import SqlCompiler
from .data import SqlFieldInfo, SqlDataSupplier, SqlTableStream
from .eval_utils import SqlEvaluationContext
from .misc import loose_table_match
from .modify_sql import embed_parameters
from .parser import SqlParser
from .. import DSLibraryException
from ..utils.df_utils import coerce_to_dataframe


def accessor_for_tables(tables, table_lookup: callable=None, table_loader: callable=None, fallback_data=None):
    """
    Generate a data supplier that will read from a collection of named pandas dataframes.
    :param tables:  A {} mapping table names to pandas.DataFrame.
    :param table_lookup:  A function to look up a real table name from a supplied table name tuple.  The default uses
                          loose_table_match() on the last table name element.
    :param table_loader:  A function that looks up a table name tuple and returns a DataFrame.
    :param fallback_data:  If no table is matched, this data is queried.
    """
    if tables is None:
        tables = {}

    class Stream(SqlTableStream):
        def __init__(self, name, use_name):
            super(Stream, self).__init__()
            self.name = name if isinstance(name, str) else name[-1]
            if table_loader:
                data = table_loader(name)
            else:
                data = tables.get(self.name)
            if data is None:
                data = fallback_data
            self.dataframe = coerce_to_dataframe(data)
            self.fields = [SqlFieldInfo(use_name, field) for field in self.dataframe]

        def supplyContext(self, context):
            self.context = context.derive(fields=self.getFields())

        def getFields(self):
            return self.fields

        def getRowIterator(self, where=None, sort=None, start=0, count=-1):
            df = self.dataframe
            if count != -1:
                df = df.iloc()[start:start + count]
            elif start:
                df = df.iloc()[start:]
            for row in df.itertuples(index=False):
                yield row

        def canLimit(self, start, count):
            return True

        def canWhere(self, where):
            # TODO support this by compiling 'where' into a pandas filter
            if where:
                return False
            return True

        def canSort(self, sort):
            # TODO support this using sort_values()
            if sort:
                return False
            return True

    class Accessor(SqlDataSupplier):
        def getTableStream(self, table_spec):
            if table_loader:
                return Stream(table_spec, table_spec[-1])
            if table_lookup:
                table_name = table_lookup(table_spec)
            else:
                table_name = loose_table_match(table_spec[-1], tables)
            if table_name in tables or fallback_data is not None:
                return Stream(table_name or table_spec[-1], table_spec[-1])

    return Accessor()


def sql_for_data_suppliers(sql, data_suppliers, ext_vars: dict=None, sql_parameters: list=None):
    """
    Run SQL against a set of DataSupplier instances, returning a DataFrame.

    :param sql:                 SQL to run.
    :param data_suppliers:      Sources of data (i.e. 'all the tables').
    :param ext_vars:            Variables referenced by the SQL.
    :param sql_parameters:      Parameters to inject into SQL that references them, i.e. with '?'.

    :return:        A DataFrame.
    """
    if sql_parameters:
        sql = embed_parameters(sql, sql_parameters)
    context = SqlEvaluationContext(data_suppliers=data_suppliers, ext_vars=ext_vars)
    select = SqlParser(sql).parse_select()
    stream = SqlCompiler().compile_select(select, context)
    columns = tuple(f.field for f in stream.getFields())
    return pandas.DataFrame(data=stream.getRowIterator(), columns=columns)


def sql_for_tables(sql, tables: dict=None, ext_vars: dict=None, table_loader=None, fallback_data=None, sql_parameters: list=None):
    """
    Run SQL against a set of tables.

    :param tables:          A {} mapping table names to dataframes.
    :param sql:             SQL to run.
    :param ext_vars:        External variables referenced by the SQL.
    :param table_loader:    A function that will return a DataFrame given a table name.
    :param fallback_data:   DataFrame to use in case a table is not matched.
    :param sql_parameters:      Parameters to inject into SQL that references them, i.e. with '?'.

    :return:    A DataFrame.
    """
    if bool(tables) + bool(table_loader) + bool(fallback_data is not None) < 1:
        raise DSLibraryException("Specify at least one of 'tables', 'table_loader' or 'fallback_data'")
    if bool(tables) + bool(table_loader) > 1:
        raise DSLibraryException("Specify at most one of 'tables' and 'table_loader'")
    if sql_parameters:
        sql = embed_parameters(sql, sql_parameters)
    src = accessor_for_tables(tables, table_loader=table_loader, fallback_data=fallback_data)
    return sql_for_data_suppliers(sql, [src], ext_vars=ext_vars)



"""
Extract useful information from parsed SQL.
"""
from dslibrary.sql.compiled import Node_Field


def find_referenced_fields(node, adder=None):
    """
    Find all fields referenced in a given expression (node).
    :param node:    Expression to search.
    :param adder:   Method which will be called with each fieldName tuple.
    :return:    A set of unique fields if adder is not given, or None.
    """
    if adder is None:
        out = set()
        find_referenced_fields(node, out.add)
        return out
    if not node:
        return
    if isinstance(node, Node_Field):
        adder(node.fieldName)
        return
    if hasattr(node, "children"):
        for sub in node.children:
            find_referenced_fields(sub, adder)


def find_groupby_fields(group_by):
    """
    Find all fields used in a GROUP BY expression.
    """
    out = set()
    if group_by:
        for lvl in group_by.levels:
            out |= find_referenced_fields(lvl[0])
    return out


def find_orderby_fields(order_by):
    return find_groupby_fields(order_by)


def find_select_fields_pre_post(fields):
    """
    For 'Select.fields', which describes what calculations to make on input fields, return:
      (0) the set of input fields used
      (1) the set of output fields produced, in the form of a mapping from output field name to input field name
        when there is a 1:1 mapping, or to None when there is not.

    A+1 as A, B, A+B as C
    input fields used: A, B
    output fields: A -> None, B -> B, C -> C
    """
    f_in = set()
    f_out = {}
    for expr, alias in (fields or []):
        f_in |= find_referenced_fields(expr)
        if isinstance(expr, Node_Field):
            f_out[(alias,) if alias else expr.fieldName] = expr.fieldName
        elif alias:
            f_out[(alias,)] = None
    return f_in, f_out


def is_wildcard_all(fields):
    """
    Test for a wildcard indicating all fields are selected.
    """
    for expr, alias in (fields or []):
        if isinstance(expr, Node_Field) and expr.fieldName[-1] == "*":
            return True
    return False



import dslibrary

from .compiler import SqlCompiler
from .eval_utils import SqlEvaluationContext
from .misc import table_name_from_table_spec, format_options_for_table
from .parser import SqlParser
from .data import SqlTableStream, SqlFieldInfo, SqlDataSupplier


def open_row_stream(resource_name, chunksize_bytes=10000000, table_format_options=None, **kwargs):
    """
    Generate a streaming row iterator that returns one row's data at a time.  The iterator
    will have a 'columns' property with column names.
    :param resource_name:   A filename or a file-like object, or the name of any resource that dslibrary is able to
        open.  A ChunkedRowIterator() can also be returned, to reference arbitrary external data.
    :param chunksize_bytes: Desired bytes per chunk for streaming.
    :param table_format_options: DataFileMetrics formatting overrides.
    :param kwargs:          Additional arguments to send to dslibrary.load_dataframe().
    :return:                A streaming row iterator.
    """
    if isinstance(resource_name, ChunkedRowIterator):
        return resource_name
    format_options = dict(table_format_options or {})
    format_options["chunksize_bytes"] = chunksize_bytes
    reader = dslibrary.load_dataframe(resource_name, format_options=format_options, **kwargs)
    rows = ChunkedRowIterator(reader)
    if hasattr(resource_name, "read"):
        rows.on_close.append(resource_name.close)
    return rows


def streaming_query(sql: str, stream_opener, chunksize_bytes=10000000, format_options=None):
    """
    Run SQL against streaming file-like data sources.  Only CSV is supported at the moment.
    :param sql:             SQL to run
    :param stream_opener:   Opens a file-like object given the name of a table.
    :param chunksize_bytes:      Approximate bytes per chunk.
    :param format_options:  A mapping from table name to a {} with DataFileMetrics format options.  Use a blank
                            key for default format options.
    :return:                A tuple with 0) column names, and 1) a row iterator for row values.
    """
    def opener(table_spec):
        table_name = table_name_from_table_spec(table_spec)
        clean_spec = (table_name,) if isinstance(table_spec, str) else table_spec[:-1] + (table_name,)
        base_stream = stream_opener(clean_spec)
        tbl_opts = format_options_for_table(format_options, table_spec)
        return open_row_stream(base_stream, chunksize_bytes=chunksize_bytes, table_format_options=tbl_opts)
    data = GenericDataSupplier(opener=opener)
    c = SqlCompiler()
    sel = SqlParser(sql).parse_select_only()
    ctx = SqlEvaluationContext(compiler=c)
    ctx.addDataSupplier(data)
    compiled = c.compile_select(sel, ctx)
    compiled.supplyContext(ctx)
    cols = tuple(field.alias or field.field for field in compiled.getFields())
    rows = compiled.getRowIterator()
    return cols, rows


class ChunkedRowIterator(object):
    """
    Turn a pandas TextFileReader instance into a row iterator.  Each iteration yields a tuple of column values for one
    row.  The 'columns' field indicates the names of each column.
    """
    def __init__(self, reader, cols=None, rows=None, closer=None):
        """
        Supply either 'reader', i.e. the result of pandas.load_csv() with a chunk_size specified, or 'cols' and 'rows'.

        :param reader:      An object with a get_chunks() and a close() method, or None.
        :param cols:        Column names.
        :param rows:        An iterable list of tuples for each row.
        :param closer:      Method to call when iteration ends.
        """
        self.reader = reader
        self.columns = cols
        self.row_gen = rows
        self.dtypes = None
        self.on_close = []
        if closer:
            self.on_close.append(closer)
        if self.row_gen is None:
            self._next_chunk()

    def _next_chunk(self):
        df = self.reader.get_chunk()
        if not self.columns:
            self.columns = list(df.columns)
            self.dtypes = list(df.dtypes)
        self.chunk_len = df.shape[0]
        self.chunk_iter = df.itertuples(index=False, name=None)

    def __iter__(self):
        if self.row_gen:
            return iter(self.row_gen)
        return self

    def __next__(self):
        while True:
            try:
                return next(self.chunk_iter)
            except StopIteration:
                self._next_chunk()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        if self.reader is not None:
            self.reader.close()
        self.reader = None
        for handler in self.on_close:
            handler()
        self.on_close.clear()


class RowIteratorTableStream(SqlTableStream):
    """
    An adapter for ChunkedRowIterator to plug it into the SQL interpreter.
    """
    def __init__(self, table_name, open_stream):
        super(SqlTableStream, self).__init__()
        self.open_stream = open_stream
        self.table_name = table_name
        stream = open_stream()
        self.fields = [SqlFieldInfo(table_name, field) for field in stream.columns]
        stream.close()

    def supplyContext(self, context):
        self.context = context.derive(fields=self.getFields())

    def getFields(self):
        return self.fields

    def getRowIterator(self, where=None, sort=None, start=0, count=-1):
        # TODO support where and other features here: pandas is faster than we are at this
        stream = self.open_stream()
        for row in iter(stream):
            yield row
        stream.close()

    def canLimit(self, start, count):
        if start or count >= 0:
            return False
        return True

    def canWhere(self, where):
        return not where

    def canSort(self, sort):
        if not sort:
            return True
        return False


class GenericDataSupplier(SqlDataSupplier):
    """
    Adapter for the SQL interpreter that supports streaming of large files.  See streaming_query().
    """
    def __init__(self, opener, db_name: str = None):
        self.opener = opener
        self.db_name = db_name

    def getTableStream(self, table_spec):
        table_name = table_spec[-1]
        db_name = table_spec[-2] if len(table_spec) > 1 else None
        if self.db_name or db_name:
            if not self.db_name or not db_name:
                return None
            if db_name.lower() != self.db_name.lower():
                return None
        return RowIteratorTableStream(table_name, lambda: self.opener(table_name))



from .misc import sql_enquote_id, loose_table_match
from .eval_utils import cmp_table_names, find_field_in_list


class SqlSelect(object):
    """
    Parsed representation of a SELECT statement.
    """
    class FromExpr(object):
        def __init__(self):
            """
            table = a () with table components, i.e. db.schema.table, or just ('table'.)
            subquery = an SqlSelect instance
            alias = alternate name
            join - a Join instance
            """
            self.table = None
            self.subquery = None
            self.alias = None
            self.join = None

        def matchesTableRef(self, table_ref):
            # check database
            db_ref = table_ref[-2] if len(table_ref) > 1 else None
            if db_ref:
                if not self.table:
                    return False
                db = self.table[-2] if len(self.table) > 1 else None
                if not cmp_table_names(db_ref, db):
                    return False
            # check table name or alias
            if self.table and cmp_table_names(table_ref[-1], self.table[-1]):
                return True
            if self.alias and cmp_table_names(table_ref[-1], self.alias):
                return True
            return False

        def __str__(self):
            return self.to_sql()

        def to_sql(self, flavor: str=None):
            if self.subquery:
                return "({0})".format(self.subquery)
            out = SqlSelect._tblspec_to_sql(self.table, flavor)
            if self.alias:
                out += " AS " + self.alias
            if self.join:
                out += " " + self.join.to_sql(flavor)
            return out

    class Join(FromExpr):
        def __init__(self):
            """
            joinType = a collection of attributes like 'left', 'right', 'inner' or 'outer'
            joinExpr = a Node_Expr instance with the join condition
            """
            SqlSelect.FromExpr.__init__(self)
            self.joinType = set()
            self.joinExpr = None

        def __str__(self):
            return self.to_sql()

        def to_sql(self, flavor: str=None):
            out = " ".join(sorted(p for p in self.joinType if p != "join")).upper()
            out += " JOIN"
            if self.subquery:
                out += " ({0})".format(self.subquery.to_sql(flavor))
            if self.table:
                out += " " + SqlSelect._tblspec_to_sql(self.table, flavor)
            if self.alias:
                out += " AS " + self.alias
            if self.joinExpr:
                out += " ON " + self.joinExpr.to_sql(flavor)
            return out

    @staticmethod
    def _tblspec_to_sql(tblspec, flavor: str=None):
        if len(tblspec) > 1 and tblspec[0]:
            return ".".join(sql_enquote_id(p, flavor) for p in tblspec if p)
        else:
            return sql_enquote_id(tblspec[0], flavor)

    def __init__(self):
        """
        Create an object representation of a SQL SELECT statement.

        from_expr = Table or table expression which is the source of data.
        where = WHERE clause, i.e. filter
        fields = what to select - this is a [] of (expression, alias)
        group_by = a GroupBy instance
        having = a Node_Expr instance
        order_by = an OrderBy instance
        start = offset in rows from the beginning
        limit = maximum number of rows to fetch
        distinct = whether to eliminate duplicate rows - a [] of column names
        union = a [] of SqlSelect instances to concatenate
        """
        self.from_expr = None
        self.where = None
        self.fields = []
        self.group_by = None
        self.having = None
        self.order_by = None
        self.start = None
        self.limit = None
        self.distinct = None
        # a list of additional selects to union with this one
        self.union = None

    def getAllTables(self):
        out = []
        tbl = self.from_expr
        while tbl:
            if tbl.subquery:
                for t in tbl.subquery.getAllTables():
                    if t not in out:
                        out.append(t)
            elif tbl.table not in out:
                out.append(tbl.table)
            tbl = tbl.join
        if self.union:
            for u in self.union:
                out += u.getAllTables()
        return out

    def map_tables(self, replacer: callable, node=None):
        """
        Visit and replace/modify all table references.
        """
        tbl = node or self.from_expr
        while tbl:
            if tbl.subquery:
                tbl.subquery.map_tables(replacer)
            else:
                tbl.table = replacer(tbl.table)
            tbl = tbl.join
        for expr in self.allExpressions():
            self._map_tables_in_expr(replacer, expr)
        if self.union:
            for u in self.union:
                u.map_tables(replacer)

    def _map_tables_in_expr(self, replacer: callable, node):
        if hasattr(node, "select_spec"):
            self.map_tables(node.select_spec)
        elif hasattr(node, "children"):
            for sub in node.children:
                self._map_tables_in_expr(replacer, sub)

    def findSubqueryByAlias(self, alias):
        tbl = self.from_expr
        while tbl:
            if tbl.subquery and tbl.alias == alias:
                return tbl.subquery
            tbl = tbl.join

    def getAliasToTableMapping(self):
        """
        Generate a mapping of table aliases to full table names.
        """
        out = {}
        tbl = self.from_expr
        while tbl:
            #NOTE: we do not collect aliases for subqueries
            if tbl.alias and tbl.table:
                out[tbl.alias] = tbl.table
            tbl = tbl.join
        return out

    def isWildcard(self):
        """
        Detect a simple "*" wildcard that passes through every field.
        """
        if len(self.fields) != 1:
            return False
        f0 = self.fields[0][0]
        if not hasattr(f0, "fieldName") or f0.fieldName[-1] != "*":
            return False
        # joins complicate the meaning of 'simple pass-through'
        if self.from_expr.join:
            return False
        return True

    def _table_ref_only(self, field_ref, get_fields_for_table):
        """
        A table was specified, we can just look up the table, and assume the field is supposed to be there.
        """
        ref_table = field_ref[:-1]
        ref_field = field_ref[-1]
        tbl = self.from_expr
        while tbl:
            if tbl.matchesTableRef(ref_table):
                if tbl.subquery:
                    return tbl.subquery.findFieldRef(field_ref[-1:], get_fields_for_table, allow_implicit_match=False)
                return tbl.table + (ref_field,)
            tbl = tbl.join

    def _implicit_table_ref(self, field_ref, get_fields_for_table):
        """
        Handle the case where there are no joins, and therefore we know which table to search.
        """
        ref_field = field_ref[-1]
        if self.from_expr.subquery:
            return self.from_expr.subquery.findFieldRef(field_ref, get_fields_for_table, allow_implicit_match=False)
        return self.from_expr.table + (ref_field,)

    def findFieldRef(self, field_ref, get_fields_for_table=None, allow_implicit_match=True):
        """
        Given the field reference tuple from a Node_Field instance (a tuple of table name/alias and field name),
        return a 3-tuple that fully describes a field, listing the proper database name and table name, followed by
        the field name.
        """
        ref_table = _normalize_table_ref(field_ref[:-1])
        ref_field = field_ref[-1]
        # if there is no 'from', then no field reference can be satisfied, by definition
        if not self.from_expr:
            return None
        # if there is only one table (and if no table was specified) the association is clear
        if not ref_table[-1] and not self.from_expr.join and allow_implicit_match:
            return self._implicit_table_ref(field_ref, get_fields_for_table)
        # if a table reference was given we can just look it up and assume the field is in that table
        if ref_table[-1]:
            return self._table_ref_only(field_ref, get_fields_for_table)
        # find table by matching ref_field, using getFieldsForTable
        tbl = self.from_expr
        while tbl:
            if tbl.subquery:
                found = tbl.subquery.findFieldRef(field_ref, get_fields_for_table, allow_implicit_match=False)
                if found:
                    return found
            elif get_fields_for_table:
                col_names = get_fields_for_table(tbl.table)
                if loose_table_match(ref_field, col_names or []):
                    return tbl.table + (ref_field,)
            tbl = tbl.join

    def _exprs_in_details(self, includeUnions=True):
        """
        Enumerate expressions outside the main body of a SELECT.
        """
        if self.group_by:
            for level in self.group_by.levels:
                yield level[0]
        if self.order_by:
            for level in self.order_by.levels:
                yield level[0]
        if self.having:
            yield self.having
        if self.start:
            yield self.start
        if self.limit:
            yield self.limit
        for u in (self.union or []) if includeUnions else []:
            for expr in u.allExpressions():
                yield expr

    def allExpressions(self, includeSubqueries=True, includeUnions=True):
        """
        Iterate through all expressions in the statement.
        """
        for field in self.fields:
            yield field[0]
        tbl = self.from_expr
        while tbl:
            if includeSubqueries and tbl.subquery:
                for expr in tbl.subquery.allExpressions():
                    yield expr
            if hasattr(tbl, "joinExpr") and tbl.joinExpr:
                yield tbl.joinExpr
            tbl = tbl.join
        if self.where:
            yield self.where
        # check everywhere else
        for expr in self._exprs_in_details(includeUnions=includeUnions):
            yield expr

    def simplify(self, context):
        self.fields = [(expr.simplify(context), alias) for expr, alias in self.fields]
        if self.where:
            self.where = self.where.simplify(context)
        if self.having:
            self.having = self.having.simplify(context)
        if self.group_by:
            self.group_by.levels = [(expr.simplify(context), order) for expr, order in self.group_by.levels]
        if self.order_by:
            self.order_by.levels = [(expr.simplify(context), order) for expr, order in self.order_by.levels]
        if self.start:
            self.start = self.start.simplify(context)
        if self.limit:
            self.limit = self.limit.simplify(context)
        if self.union:
            self.union = [sel.simplify(context) for sel in self.union]
        # TODO also simplify the sub-expressions within 'from_expr'
        return self

    def ok_to_cache_as_subquery(self, context):
        """
        Test whether all the fields used in the expression can be satisfied by the given context.

        If so, it is assumed that nothing varies, and caching can be used.
        """
        for expr in self.allExpressions(includeSubqueries=False, includeUnions=True):
            if not self._node_ok_to_cache(expr, context):
                return False
        return True

    @staticmethod
    def _node_ok_to_cache(node, context):
        if hasattr(node, "fieldName"):
            f = node.fieldName
            if find_field_in_list(f, context) == -1:
                return False
            return True
        if hasattr(node, "children"):
            for child in node.children:
                if not SqlSelect._node_ok_to_cache(child, context):
                    return False
        return True

    def __str__(self):
        return self.to_sql()

    def to_sql(self, flavor: str=None):
        """
        Convert a SELECT to SQL.
        """
        # limit & offset do not work with mssql
        use_fetch = False
        use_top = False
        if flavor == "mssql":
            if self.limit and self.start is None:
                use_top = True
            elif self.limit or self.start:
                use_fetch = True
        # build...
        out = ["SELECT"]
        if use_top:
            out += ["TOP", self.limit.to_sql(flavor)]
        if self.distinct is not None:
            if len(self.distinct):
                out.append("DISTINCT(%s)" % ", ".join(map(lambda f: sql_enquote_id(f, flavor), self.distinct)))
            else:
                out.append("DISTINCT")
        f_spec = []
        for f in self.fields:
            part = f[0].to_sql(flavor)
            if f[1]:
                part += " AS " + sql_enquote_id(f[1], flavor)
            f_spec.append(part)
        out.append(", ".join(f_spec))
        if self.from_expr:
            out.append("FROM " + self.from_expr.to_sql(flavor))
        if self.where:
            out.append("WHERE {0}".format(self.where.to_sql(flavor)))
        if self.group_by:
            out.append("GROUP BY {0}".format(self.group_by.to_sql(flavor)))
        if self.having:
            out.append("HAVING {0}".format(self.having.to_sql(flavor)))
        if self.order_by:
            out.append("ORDER BY {0}".format(self.order_by.to_sql(flavor)))
        if use_top:
            # 'top' clause already emitted, no post-clause to add
            pass
        elif use_fetch:
            if not self.order_by:
                raise Exception(f"engine type {flavor}: cannot use offset without an order by")
            out += ["OFFSET", self.start.to_sql(flavor), "ROWS"]
            if self.limit:
                out += ["FETCH", "FIRST", self.limit.to_sql(flavor), "ROWS", "ONLY"]
        else:
            if self.start:
                out.append("OFFSET {0}".format(self.start.to_sql(flavor)))
            if self.limit:
                out.append("LIMIT {0}".format(self.limit.to_sql(flavor)))
        if self.union:
            out.append("UNION")
            out.append(" UNION ".join(u.to_sql(flavor) for u in self.union))
        return " ".join(out)


def _normalize_table_ref(table_ref):
    """
    Make sure a table reference has both a database name and a table name.
    """
    if len(table_ref) < 2:
        return (None,) + table_ref
    return table_ref



from .compiled import uses_aggregation, Node_Field, SqlOrderBy
from .data import SqlTableStream, MemoryDataSupplier, SqlFieldInfo
from .eval_utils import find_field_in_list
from .exceptions import SqlException, SqlException_DataSourceNotFound
from .parser import expand_field_spec
from .select_stmt import SqlSelect
from .streams import _JoinMerger, where_and_where, test_sql_filter, _FieldCalc, Union, WhereAndLimit, SortableStream


class SqlCompiler(object):
    """
    Through an instance of this class, a parser can generate any kind of data stream, i.e. an instance of
    SqlTableStream that performs all sorts of joining, aggregation, filtering, sorting and so on.
    """

    def data_source(self, table_spec):
        assert isinstance(table_spec, tuple)
        """
        Get the lowest level of data stream.  Data will be pulled from one of the
        data suppliers provided by the context.
        """

        class DataSource(SqlTableStream):
            def __init__(self):
                super(DataSource, self).__init__()
                self.source = None

            def supplyContext(self, context):
                self.source = context.findDataSource(table_spec)
                if not self.source:
                    raise SqlException_DataSourceNotFound("not found: {0}".format(".".join(table_spec)))
                self.source.supplyContext(context)
                self.context = context.derive(fields=self.getFields())
                self.upstream = self.source

            def getFields(self):
                return self.source.getFields()

            def getRowIterator(self, where=None, sort=None, start=0, count=-1):
                return self.source.getRowIterator(where, sort, start, count)

            def canSort(self, sort):
                return self.source.canSort(sort)

            def canWhere(self, where):
                return self.source.canWhere(where)

            def canLimit(self, start, count):
                return self.source.canLimit(start, count)

        return DataSource()

    def distinctify(self, stream, fields=None):
        """
        Filter out duplicate values from a stream.
        :param stream:   Stream to apply to.
        :param fields:   List of field specifiers - an iterable of strings or tuples.
        :return:    New stream.
        """

        class Distinct(SqlTableStream):
            def __init__(self):
                SqlTableStream.__init__(self, stream)

            def getRowIterator(self, where=None, sort=None, start=0, count=-1):
                distinct = set()
                ctx = self.context.derive(fields=self.getFields())

                def value_to_make_distinct(row):
                    if not fields:
                        return row
                    out = []
                    for f in fields:
                        idx = find_field_in_list(f, ctx)
                        if idx != -1:
                            out.append(row[idx])
                    return tuple(out)

                for row in stream.getRowIterator(where, sort, start, count):
                    v = value_to_make_distinct(row)
                    if v not in distinct:
                        distinct.add(v)
                        yield row

        return Distinct()

    def join(self, stream_l, stream_r, join_type, join_expr):
        """
        Merge two streams into one.
        :param stream_l:
        :param stream_r:
        :param join_type:
        :param join_expr:
        :return:
        """
        return _JoinMerger(stream_l, stream_r, join_type=join_type, join_expr=join_expr, compiler=self)

    def support_where_and_limit(self, stream, force_post_filter=False):
        """
        Add filtering capability to a stream.
        """
        out = WhereAndLimit(stream, force_post_filter)
        if stream.context:
            out.supplyContext(stream.context)
        return out

    def support_sort(self, stream):
        """
        Add sorting support to a stream.
        """
        return SortableStream(stream)

    def apply_where(self, stream, where_condition, force_post_filter=False):
        """
        Filter a stream.
        """
        class Where(SqlTableStream):
            def __init__(self):
                SqlTableStream.__init__(self, stream)
                if stream.context:
                    self.supplyContext(stream.context)

            def getRowIterator(self, where=None, sort=None, start=0, count=-1):
                combined_where = where_and_where(where, where_condition)
                if not force_post_filter and stream.canWhere(combined_where):
                    return stream.getRowIterator(combined_where, sort, start, count)

                def gen():
                    ctx0 = self.context.derive(fields=self.getFields())
                    for row in stream.getRowIterator(None, sort, start, count):
                        ctx = ctx0.derive(record=row)
                        if test_sql_filter(combined_where, ctx):
                            yield row

                return gen()

        return Where()

    def apply_sort(self, stream, sort_spec):
        """
        Sort a stream.
        """
        cmp = self

        class Sort(SqlTableStream):
            def __init__(self):
                SqlTableStream.__init__(self, stream)
                if stream.context:
                    self.supplyContext(stream.context)

            def getRowIterator(self, where=None, sort=None, start=0, count=-1):
                combined_sort = sort or sort_spec
                if stream.canSort(combined_sort):
                    return stream.getRowIterator(where, combined_sort, start, count)
                return cmp.get_full_row_iterator(stream, where, combined_sort, start, count)

            def canSort(self, sort):
                return True

            def canLimit(self, start, count):
                return True

        return Sort()

    def apply_limit(self, stream, apply_start=0, apply_count=-1):
        """
        Apply limits to a stream.
        """
        cmp = self

        class Limits(SqlTableStream):
            def __init__(self):
                SqlTableStream.__init__(self, stream)

            def getRowIterator(self, where=None, sort=None, start=0, count=-1):
                limits = combine_limits((apply_start, apply_count), (start, count))
                return cmp.get_full_row_iterator(stream, where, sort, limits[0], limits[1])

            def canLimit(self, start, count):
                return True

        return Limits()

    def get_full_row_iterator(self, stream, where=None, sort=None, start=0, count=-1):
        """
        Fills in whatever capabilities a stream is missing.
        """
        full = stream
        if (where and not full.canWhere(where)) or ((start or count >= 0) and not full.canLimit(start, count)):
            assert stream.context
            full = self.support_where_and_limit(full)
            full.supplyContext(stream.context)
        if sort and not full.canSort(sort):
            assert stream.context
            full = self.support_sort(full)
            full.supplyContext(stream.context)
        return full.getRowIterator(where, sort, start, count)

    def calculate_or_select_fields(self, stream, field_mappings, **kwargs):
        """
        Modify the list of fields in a stream.  Each entry in fieldMappings is a tuple with the following:
          [0]: the expression to evaluate
          [1]: the new alias to give the field, if any
        """
        # create the stream that modifies fields
        return _FieldCalc(stream, compiler=self, field_mappings=field_mappings, **kwargs)

    def generate_aggregation_grouper(self, group_by):
        """
        Perform aggregation.  The supplied 'aggregationSpec' is an iterable with either a SqlAggregator
        instance or None for each field position.
        """
        def grouper(context):
            if not group_by:
                return 0
            return tuple(level[0].eval(context) for level in group_by.levels)
        return grouper

    def union(self, streams):
        """
        Concatenate streams.
        """
        return Union(streams)

    def add_table_alias(self, stream, new_alias):
        """
        Adds a table alias to a stream.
        """
        return AliasAdder(stream, new_alias)

    def compile_select(self, select, context=None, subquery=False):
        """
        Generate a table stream, given a parsed select statement.

        :param select:  an instance of SqlSelect, i.e. from SqlParser.parseSelect()
        :param context: an instance of SqlUtil.SqlEvaluationContext()
        :param subquery: hint: this is a subquery
        """
        assert isinstance(select, SqlSelect)

        def find_table(from_expr):
            if from_expr.subquery:
                s = self.compile_select(from_expr.subquery, context)
            else:
                s = self.data_source(from_expr.table)
            if from_expr.alias:
                s = self.add_table_alias(s, from_expr.alias)
            return s

        # generate stream using FROM expression
        if not select.from_expr:
            stream = self._compile_select__single_row(select, context)
        else:
            stream = self._compile_select__composite_from(select, subquery, find_table, context)
        # qualify with WHERE, etc.
        stream = self._compile_select__qualify(select, stream, context)
        # result is 'stream'
        if context:
            stream.supplyContext(context)
        return stream

    def _value_or_eval(self, spec, context, default):
        if spec is None:
            return default
        return spec.eval(context)

    def _compile_select__qualify(self, select, stream, context):
        # WHERE
        if select.where:
            stream = self.apply_where(stream, select.where)
        # HAVING
        if select.having:
            stream = self.apply_where(stream, select.having, force_post_filter=True)
        # ORDER BY
        if select.order_by:
            stream = self.apply_sort(stream, select.order_by)
        # START/LIMIT
        start = self._value_or_eval(select.start, context, 0)
        limit = self._value_or_eval(select.limit, context, -1)
        if start or limit not in {0, -1}:
            stream = self.apply_limit(stream, start, limit or -1)
        # UNION
        if select.union:
            to_union = [stream] + [self.compile_select(sel, context) for sel in select.union]
            stream = self.union(to_union)
        # DISTINCT
        if select.distinct is not None:
            stream = self.distinctify(stream, fields=select.distinct or None)
        return stream

    def _compile_select__single_row(self, select, context):
        # single row
        single_row = {}
        for field in select.fields:
            expr, alias = field
            value = expr.eval(context)
            if not alias:
                alias = str(expr)
            single_row[alias] = value
        mem = MemoryDataSupplier(data={"T": [single_row]})
        return mem.getTableStream(("T",))

    def _compile_select__composite_from(self, select, subquery, find_table, context):
        # FROM/JOIN
        alias_to_table = select.getAliasToTableMapping()
        stream = find_table(select.from_expr)
        join = select.from_expr.join
        d_from_u = subquery
        while join:
            tbl = find_table(join)
            stream = self.join(stream, tbl, join.joinType, join.joinExpr)
            join = join.join
        # GROUP BY
        upstream_sort = None
        grouper = None
        if uses_aggregation(select):
            # pre-sort the stream based on the grouping
            if select.group_by:
                upstream_sort = select.group_by
            # generate grouping function to aggregate results
            grouper = self.generate_aggregation_grouper(select.group_by)
            d_from_u = True
        # FIELDS
        if select.isWildcard() and not grouper:
            # no modifications or field calculations or aggregation--don't bother wrapping the stream
            return stream
        field_list = _expand_wildcards(select, alias_to_table, context)
        table_to_alias = {alias_to_table[k]: k for k in alias_to_table}
        return self.calculate_or_select_fields(
            stream, field_list, alias_to_table=alias_to_table,
            table_to_alias=table_to_alias, aggregation_grouper=grouper,
            downstream_visible_from_upstream=d_from_u, upstream_sort=upstream_sort
        )


class AliasAdder(SqlTableStream):
    def __init__(self, upstream, new_alias):
        SqlTableStream.__init__(self, upstream)
        self.fields = []
        self.new_alias = new_alias

    def remove_alias(self, node):
        if isinstance(node, Node_Field):
            field_name = node.fieldName
            if len(field_name) > 1 and field_name[-1] == self.new_alias:
                return Node_Field(field_name[-1:])
            else:
                return node
        elif hasattr(node, "children"):
            new_node = node.clone()
            new_node.children = [self.remove_alias(sub) for sub in node.children]
            return new_node
        else:
            return node

    def supplyContext(self, context):
        SqlTableStream.supplyContext(self, context)
        ff = []
        for f in self.upstream.getFields():
            ff.append(SqlFieldInfo(table=f.table, field=f.field, database=f.database, alias=f.alias,
                                           table_alias=self.new_alias))
        self.fields = ff
        self.context = context.derive(fields=self.fields)

    def getFields(self):
        return self.fields

    def getRowIterator(self, where=None, sort=None, start=0, count=-1):
        # remove the alias from where and sort
        if where:
            where = self.remove_alias(where)
        if sort:
            s2 = SqlOrderBy()
            for lvl in sort.levels:
                s2.addLevel(self.remove_alias(lvl[0]), lvl[1])
            sort = s2
        return self.upstream.getRowIterator(where, sort, start, count)


def combine_limits(limits1, limits2):
    """
    Apply one set of (start, count) to another.

    Limits1 are applied first, then limits2 apply next.
    """
    start = limits1[0]
    count = limits1[1]
    if limits2[0] > 0:
        start2 = limits2[0]
        if count != -1 and start2 > count:
            count = 0
        else:
            start += start2
            if count != -1:
                count -= start2
    if limits2[1] != -1:
        count2 = limits2[1]
        if count2 < count or count == -1:
            count = count2
    return (start, count)


class _WildcardExpander(object):
    def __init__(self, context, table_aliases):
        self.context = context
        self.table_aliases = table_aliases
        self.out = []

    def get_fields_for_table(self, table_spec):
        src = self.context.findDataSource(table_spec)
        if src:
            return [field.field for field in src.getFields()]
        return []

    def add(self, field, expanded):
        expanded = list(expanded)
        if len(expanded) == 1:
            # single field - we can preserve the alias
            if field[-1]:
                self.out.append((Node_Field(expanded[0]), field[1]))
            return
        for f in expanded:
            if f[-1]:
                self.out.append((Node_Field(f), None))

    def find(self, select_spec):
        found_wild = False
        for f in select_spec.fields:
            if not(isinstance(f[0], Node_Field) and f[0].isWildcard()):
                self.out.append(f)
                continue
            if not self.context:
                raise SqlException("no context provided, cannot expand wildcards")
            found_wild = True
            expanded = expand_field_spec(
                select_spec, f[0].fieldName, self.get_fields_for_table, alias_to_table=self.table_aliases
            )
            # custom extension to SQL: a 'non-greedy' wildcard that lets you pull in "all other fields"
            if f[0].fieldName[-1].endswith("?"):
                # the "?" prevents duplicate field names from being selected
                prev_fields = {o[0].fieldName[-1] if isinstance(o[0], Node_Field) else o[1] for o in
                               self.out}
                expanded = filter(lambda x: x[-1] not in prev_fields, expanded)
            # add to results
            self.add(f, expanded)
        if found_wild:
            return self.out
        return select_spec.fields


def _expand_wildcards(select_spec, table_aliases, context):
    """
    Expand "*" and "table.*" references in a field list.
    Returns an expanded field list.
    """
    assert isinstance(select_spec, SqlSelect)
    exp = _WildcardExpander(context, table_aliases)
    return exp.find(select_spec)



"""
Tokenize SQL.  Supports a wide range of SQL grammars.
"""


class Token:
    # token types
    PUNCT = 'p'
    STRING = 's'
    KEYWORD = 'k'
    REAL = 'r'
    INTEGER = 'i'
    PLACEHOLDER = '?'

    def __init__(self, value, type, pos, len, etc=None):
        self.value = value
        self.type = type
        self.pos = pos
        self.len = len
        self.etc = etc

    def isLiteral(self):
        return self.type in {self.STRING, self.REAL, self.INTEGER, self.PLACEHOLDER}

    def isPunct(self, p):
        return self.type in {self.PUNCT, self.PLACEHOLDER} and self.value == p

    def isKeyword(self, kwd):
        return self.type == self.KEYWORD and self.value.lower() == kwd.lower()


class SqlTokenizer(object):
    WS_CHAR = " \t\r\n\f\b"
    ID_CHAR_START = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz_"
    DIGIT_CHAR = "0123456789"
    ID_CHAR = ID_CHAR_START + DIGIT_CHAR + "."
    NUMBER_CHAR = DIGIT_CHAR + "."
    NON_PUNCT = WS_CHAR + ID_CHAR + NUMBER_CHAR + "'"
    VALID_T_SQL_START = ID_CHAR_START

    def __init__(self, input: str, allowed_operators: set, allow_placeholders: bool=False):
        self.input = input
        self.pos = 0
        self.len = len(input)
        self.tokens = []
        self.allowed_operators = allowed_operators
        self.allow_placeholders = allow_placeholders

    def _comment1(self):
        while self.pos < self.len and self.input[self.pos] != '\n':
            self.pos += 1
        if self.pos < self.len and self.input[self.pos] == '\n':
            self.pos += 1

    def _comment2(self):
        self.pos += 1
        while self.pos < self.len - 1 and self.input[self.pos] != '*' and self.input[self.pos + 1] != '/':
            self.pos += 1
        if self.pos < self.len - 1 and self.input[self.pos] == '*' and self.input[self.pos + 1] == '/':
            self.pos += 2

    def _string_literal(self, c):
        quote = c
        p0 = self.pos
        while self.pos < self.len:
            c = self.input[self.pos]
            if c != quote:
                self.pos += 1
            else:
                if self.pos + 1 < self.len and self.input[self.pos + 1] == quote:
                    self.pos += 2
                else:
                    break
        value = self.input[p0:self.pos].replace(quote + quote, quote)
        if self.pos < self.len and self.input[self.pos] == quote:
            self.pos += 1
        self.tokens.append(Token(value, Token.STRING if quote == "'" else Token.KEYWORD, p0 - 1, self.pos - p0 + 1, etc=c))

    def _keyword(self):
        p0 = self.pos - 1
        while self.pos < self.len and self.input[self.pos] in self.ID_CHAR:
            self.pos += 1
        value = self.input[p0:self.pos]
        if "." in value:
            # keywords separated by '.' are a special case, i.e. for qualified table names
            for n_value, value in enumerate(value.split(".")):
                if n_value:
                    self.tokens.append(Token(".", Token.PUNCT, p0, 1))
                    p0 += 1
                if value:
                    self.tokens.append(Token(value, Token.KEYWORD, p0, len(value)))
                    p0 += len(value)
        else:
            self.tokens.append(Token(value, Token.KEYWORD, p0, len(value)))

    def _tsql_id(self):
        p0 = self.pos - 1
        while self.pos < self.len and self.input[self.pos] != ']' or (self.pos+1 < self.len and self.input[self.pos+1] == ']'):
            self.pos += 1
        value = self.input[p0+1:self.pos]
        value = value.replace("[[", "[").replace("]]", "]")
        self.tokens.append(Token(value, Token.KEYWORD, p0, len(value)))
        if self.pos < self.len and self.input[self.pos] == ']':
            self.pos += 1

    def _numeric(self, c):
        # numeric literal
        p0 = self.pos - 1
        if c == '0' and self.pos < self.len and self.input[self.pos] in ['x', 'X']:
            self._hex(p0)
            return
        rtype = Token.REAL if c == '.' else None
        if c in self.DIGIT_CHAR:
            rtype = Token.INTEGER
        while self.pos < self.len and self.input[self.pos] in self.DIGIT_CHAR:
            self.pos += 1
            rtype = rtype or Token.INTEGER
        if self._detect_real_exp(rtype):
            rtype = Token.REAL
        if rtype == Token.REAL:
            v_raw = self.input[p0:self.pos]
            if v_raw == ".":
                self.tokens.append(Token(v_raw, Token.PUNCT, p0, self.pos - p0))
            else:
                value = float(v_raw)
                self.tokens.append(Token(value, rtype, p0, self.pos - p0))
        elif rtype == Token.INTEGER:
            s = self.input[p0:self.pos]
            value = int(s)
            self.tokens.append(Token(value, rtype, p0, self.pos - p0))
        else:
            self.tokens.append(Token(".", Token.PUNCT, p0, 1))

    def _detect_real_exp(self, rtype):
        """
        Presence of "." or valid exponent means we have a float.
        """
        found = False
        if self.pos < self.len and self.input[self.pos] == ".":
            if rtype == Token.INTEGER:
                found = True
            self.pos += 1
            while self.pos < self.len and self.input[self.pos] in self.DIGIT_CHAR:
                found = True
                self.pos += 1
        if self.pos < self.len and self.input[self.pos] in ['e', 'E']:
            p1 = self.pos
            self.pos += 1
            if self.pos < self.len and self.input[self.pos] in ['+', '-']:
                self.pos += 1
            while self.pos < self.len and self.input[self.pos] in self.DIGIT_CHAR:
                self.pos += 1
                p1 = self.pos
                found = True
            self.pos = p1
        return found

    def _hex(self, p0):
        self.pos += 1
        while self.pos < self.len and self.input[self.pos] in self.DIGIT_CHAR + "abcdefABCDEF":
            self.pos += 1
        value = int(self.input[p0 + 2:self.pos], 16)
        self.tokens.append(Token(value, Token.INTEGER, p0, self.pos - p0, etc="hex"))

    def _punct(self):
        p0 = self.pos - 1
        if self.pos < self.len and self.input[self.pos] not in self.NON_PUNCT and \
                self.input[p0:self.pos + 1] in self.allowed_operators:
            self.pos += 1
        v = self.input[p0:self.pos]
        if v == "%" and self.input[p0:self.pos+1] == "%s" and self.allow_placeholders:
            self.tokens.append(Token("%s", Token.PLACEHOLDER, p0, 2))
            self.pos += 1
        elif self.allow_placeholders and v == "?":
            self.tokens.append(Token(v, Token.PLACEHOLDER, p0, 1))
        else:
            self.tokens.append(Token(v, Token.PUNCT, p0, self.pos-p0))

    def run(self):
        while self.pos < self.len:
            c = self.input[self.pos]
            self.pos += 1
            if c in self.WS_CHAR:
                # ignore whitespace
                pass
            elif c == '-' and self.pos < self.len and self.input[self.pos] == '-':
                self._comment1()
            elif c == '/' and self.pos < self.len - 1 and self.input[self.pos] == '*':
                self._comment2()
            elif c in ("'", '"', '`'):
                self._string_literal(c)
            elif c in self.ID_CHAR_START:
                if c == 'N' and self.pos < self.len and self.input[self.pos] == "'":
                    # sql server format:
                    self.pos += 1
                    self._string_literal("'")
                else:
                    self._keyword()
            elif c in self.NUMBER_CHAR:
                self._numeric(c)
            elif c == '[' and self.pos < self.len and self.input[self.pos] in self.VALID_T_SQL_START and (self.pos <= 1 or self.input[self.pos-2] != '*'):
                self._tsql_id()
            else:
                self._punct()
        self._combine_compound()
        return self.tokens

    def _combine_compound(self):
        # combine compound tokens
        compound = [
            ["is", "not"],
            ["not", "in"],
            ["not", "like"],
            ["not", "ilike"],
            ["not", "match"],
            ["not", "regexp"],
            ["not", "glob"]
        ]
        for p in range(0, len(self.tokens) - 1):
            if self.tokens[p].type == Token.KEYWORD and self.tokens[p].value.lower() in ["is", "not"]:
                for cp in compound:
                    if str(self.tokens[p].value).lower() == cp[0] and str(self.tokens[p + 1].value).lower() == cp[1]:
                        self.tokens[p].value = cp[0] + " " + cp[1]
                        self.tokens.pop(p + 1)
        return self.tokens



"""
Modifications to SQL.
"""
import re

from .parser import SqlParser
from .select_stmt import SqlSelect
from .compiled import Node_Literal, Node_Field, Node
from .tokenize import SqlTokenizer, Token
from .misc import enquote_string_literal, sql_enquote_id
from .eval_utils import OPS_ALL


def impose_sql_limit(sql: str, limit: int=None, flavor: str=None):
    """
    Inject a limit on some SQL.  Also reformats to the selected flavor/dialect.
    :param sql:  SQL to modify.  Must be a SELECT statement.
    :param limit: Limit to insert.
    :param flavor:  Language variant (see sql_enquote_id()).
    """
    select = SqlParser(sql, allow_any_function=True).parse_select()
    if not select:
        # not a SELECT statement
        return sql

    def apply(sel):
        use_lim = limit
        if sel.limit:
            use_lim = min(limit, sel.limit.eval(None))
        sel.limit = Node_Literal(use_lim)

    if limit:
        if select.union:
            # limit is applied to last union
            apply(select.union[-1])
        else:
            apply(select)
    return select.to_sql(flavor)


def limit_sql_cursor_rows(base_cursor, max_rows: int=None):
    """
    Prevent a DBI cursor from returning more than a given number of rows.

    :param base_cursor:  A DBI cursor, or a jdbc.Runner() instance, which wraps such a connection/cursor.
    """
    class MaxRowsCursor(object):
        def __init__(self):
            self.description = self._descr()
            self._remaining = None if max_rows is None else max(max_rows, 0)

        def _cursor(self):
            if hasattr(base_cursor, "cursor"):
                return base_cursor.cursor
            return base_cursor

        def _descr(self):
            c = self._cursor()
            if c is not None and hasattr(c, "description"):
                return c.description

        def _available(self, n: int):
            if self._remaining is None:
                return n
            if n >= self._remaining:
                return self._remaining
            return n

        def _use(self, n: int):
            if self._remaining is None:
                return
            self._remaining -= n
            if self._remaining == 0:
                self.close()

        def start(self):
            r = None
            if hasattr(base_cursor, "start"):
                r = base_cursor.start()
                if not self.description:
                    self.description = self._descr()
            return r

        def stop(self):
            self.close()

        def close(self):
            if hasattr(base_cursor, "stop"):
                base_cursor.stop()
            elif hasattr(base_cursor, "close"):
                base_cursor.close()
            self._remaining = 0

        def fetchone(self):
            n = self._available(1)
            if n:
                r = self._cursor().fetchone()
                if not r:
                    self.close()
                    self._remaining = 0
                self._use(n)
                return r

        def fetchall(self):
            if self._remaining is None:
                return self._cursor().fetchall()
            use = self._remaining
            r = self._cursor().fetchmany(use)
            self._use(use)
            return r

        def fetchmany(self, n: int):
            if self._remaining is None:
                return self._cursor().fetchmany(n)
            use = self._available(n)
            r = self._cursor().fetchmany(use)
            self._use(use)
            return r

        def __next__(self):
            r = self.fetchone()
            if not r:
                raise StopIteration
            return r

        def __iter__(self):
            return self

    return MaxRowsCursor()


def looks_like_sql(sql: str):
    """
    Differentiate between SQL and other strings.
    """
    if not sql:
        return False
    # any sql comment implies sql
    if "/*" in sql and "*/" in sql:
        return True
    # sql verbs
    if re.match(r'^\s*(with|select|insert|create|describe)\s+[^\s].*', sql, re.IGNORECASE | re.DOTALL):
        return True
    return False


def just_a_table_name(sql: str, interpret_select: bool=True):
    """
    Detect a string which is just a table name, i.e. "table", or SQL that only references a table with no modification,
    i.e. "select * from table".
    """
    if not sql or not isinstance(sql, str):
        return
    if looks_like_sql(sql):
        if not interpret_select:
            return
        m = re.match(r'^(?:\s+|/\*.*?\*/)*select\s+\*\s+from\s("[^\n]*?"|`[^\n]*?`|[^\s]+)\s*$', sql, re.DOTALL | re.I)
        if m:
            name = m.group(1)
            if len(name) >= 2 and name[0] in ('"', "`") and name[0] == name[-1]:
                name = name[1:-1]
            return name
        return
    # some strings are not likely to be table names
    if "\n" in sql or len(sql) > 255:
        return
    return sql


def _rebuild_from_tokens(sql, draw_placeholder):
    tokens = SqlTokenizer(sql, OPS_ALL, allow_placeholders=True).run()
    bits = []
    pos = 0
    n_param = 0
    for token in tokens:
        if token.type != Token.PLACEHOLDER:
            continue
        if token.pos > pos:
            bits.append(sql[pos:token.pos])
        s_token = draw_placeholder(token, n_param) or "NULL"
        bits.append(s_token)
        n_param += 1
        pos = token.pos + token.len
    bits.append(sql[pos:])
    return "".join(bits)


def embed_parameters(sql: str, parameters: (list, tuple)):
    """
    Replace placeholders in SQL with a given set of parameters.
    :param sql:                 SQL with placeholders '%s' or '?'.
    :param parameters:          An iteration of values.
    :return:        Filled-out SQL.
    """
    def draw_placeholder(token, n_param):
        if parameters and n_param < len(parameters):
            v = parameters[n_param]
            if isinstance(v, str):
                return enquote_string_literal(v)
            else:
                return str(v)
    return _rebuild_from_tokens(sql, draw_placeholder)


def reformat_placeholders(sql: str, flavor: str=None):
    """
    Reformat placeholders in SQL to match the target database engine.
    """
    if "?" not in sql and "%s" not in sql:
        return sql
    flavor_ph = "%s"
    def draw_placeholder(token, n_param):
        return flavor_ph
    return _rebuild_from_tokens(sql, draw_placeholder)


def apply_default_table_name(sql: str, use_table: str, default_table_names: tuple=("table",), flavor: str=None):
    """
    SQL can be written using a default table name, and then that default table can be replaced with an actual table
    name.
    """
    # no table supplied: return unchanged SQL
    if not use_table:
        return sql
    # table name only?  Return the supplied table name instead.
    tn_only = just_a_table_name(sql)
    if tn_only:
        if tn_only in default_table_names:
            if looks_like_sql(sql):
                return "select * from " + sql_enquote_id(use_table, mode=flavor)
            return use_table
        return sql
    sel = SqlParser(sql, allow_any_function=True).parse_select()
    def replacer(table_spec):
        if len(table_spec) == 1 and table_spec[-1] in default_table_names:
            return (use_table,)
        return table_spec
    sel.map_tables(replacer)
    return sel.to_sql(flavor)


class BuildSqlSelect(object):
    def __init__(self, sql: str=None):
        """
        Begin building a SELECT, optionally starting from the given SQL.
        """
        if sql:
            self.statement = SqlParser(sql).parse_select_only()
        else:
            self.statement = SqlSelect()

    def distinct(self):
        """
        Add the 'distinct' clause.
        """
        self.statement.distinct = []
        return self

    def field(self, expr, alias: str = None):
        """
        Add a field to return from the SELECT.
        """
        if self.statement.fields is None:
            self.statement.fields = []
        if isinstance(expr, str):
            self.statement.fields.append((Node_Field((expr,)), alias))
        elif isinstance(expr, (tuple, list)):
            self.statement.fields.append((Node_Field(tuple(part) for part in expr), alias))
        elif isinstance(expr, Node):
            self.statement.fields.append((expr, alias))
        else:
            raise Exception(f"misuse of field(), expr={expr}")
        return self

    def set_from(self, expr, alias: str=None):
        """
        Set the source table.
        """
        if isinstance(expr, str):
            self.statement.from_expr = SqlSelect.FromExpr()
            self.statement.from_expr.table = (expr,)
        elif isinstance(expr, (tuple, list)):
            self.statement.from_expr = SqlSelect.FromExpr()
            self.statement.from_expr.table = tuple(expr)
        else:
            raise Exception(f"misuse of set_from(), expr={expr}")
        self.statement.from_expr.alias = alias
        return self

    def where(self, expr):
        """
        Specify WHERE clause.
        """
        if isinstance(expr, str):
            self.statement.where = SqlParser(expr).parse_expr()
        elif isinstance(expr, Node):
            self.statement.where = expr
        else:
            raise Exception(f"misuse of where(), expr={expr}")
        return self

    def limit(self, expr):
        """
        Specify a limit.
        """
        if isinstance(expr, int):
            self.statement.limit = Node_Literal(expr)
        elif isinstance(expr, Node):
            self.statement.limit = expr
        else:
            raise Exception(f"misuse of limit(), expr={expr}")
        return self


